# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""FactsManager_FactInc"""

from nark.items.fact import UntilTimeStops

__all__ = (
    'FactsManager_FactInc',
)


# SYNC_ME: Keep FactsManager_FactDec and FactsManager_FactInc synced.
#   That, or write a generic base class, but that sounds painfuller to maintain.
#   then simply running meld as necessary:
#
#       meld facts_mgr_fact_dec.py facts_mgr_fact_inc.py &
class FactsManager_FactInc(object):
    """"""
    def jump_fact_inc(self):
        """"""
        def _jump_fact_inc():
            # Check first if we've reached the ending of all time.
            is_final_fact = (self.curr_index == (len(self.curr_group) - 1))
            if (
                is_final_fact
                and not self.curr_fact.end
                and self.curr_group.until_time_stops
            ):
                return None
            # Each facts group represents a contiguous block of time that is
            # being managed by dob, and to ignore anything in the store from
            # that time. Find either the next fact under dob's control,
            # or one from the store (that we'll wire into dob in a later step).
            next_fact = fetch_next_fact()
            # Each fact's next_fact and prev_fact links are not wired until
            # we inspect them here for gaps -- meaning, we fill in the gap facts
            # reactively, as the user traverses the facts in the carousel. Now.
            next_fact = fill_gap_until(next_fact)
            # If we've reached the ending of all time, there's nothing following.
            if next_fact is None:
                return None
            # Just a sanity check. Note that we use fact.times for >, otherwise
            # sorty_tuple is used, which also sorts by PK, which masks 2 facts
            # at same time but with unique IDs.
            # - This also shows that we update self.curr_index already, and
            #   maybe self.curr_group, but not self.curr_fact (so the state
            #   is outta sorts).
            self.controller.affirm(self.curr_fact.end <= next_fact.start)
            # We (re)wired the fact to the group earlier; now rewire the group.
            self.fulfill_jump(next_fact, reason='fact-inc')
            # See if we've identified the boundary of the known factiverse.
            if (next_fact.end is None) or (next_fact.end is UntilTimeStops):
                self.controller.affirm(self.curr_group.until_time_stops)
            self.controller.client_logger.debug('\n- next: {}'.format(next_fact.short))
            return next_fact

        # ***

        def fetch_next_fact():
            # Check curr_fact.curr_index, and not curr_fact.has_next_fact,
            # because the facts may not all be wired yet.
            self.curr_index < (len(self.curr_group) - 1)
            if self.curr_index < (len(self.curr_group) - 1):
                return from_curr_group_increment()
            else:
                # Choose between fact from store or next group (if any).
                return from_next_group_or_store()

        # ***

        def from_curr_group_increment():
            self.curr_index += 1
            next_from_curr_group = self.curr_group[self.curr_index]
            return next_from_curr_group

        # ***

        def curr_group_add_next(next_fact):
            self.controller.affirm(self.curr_fact.next_fact is None)
            self.curr_group_add(next_fact)
            if self.curr_group[self.curr_index] is not next_fact:
                self.curr_index += 1
            self.controller.affirm(self.curr_group[self.curr_index] is next_fact)
            return next_fact

        # ***

        def from_next_group_or_store():
            next_from_store = self.fetch_next_from_store()
            if overlaps_current_fact(next_from_store):
                return fix_overlapping_fact(next_from_store)
            if exists_in_between_groups(next_from_store):
                # Add the fact to the group, but do not wire yet,
                # as we may add a gap fact, if necessary.
                return curr_group_add_next(next_from_store)
            # No facts between this group and next group;
            # or there are no facts from store
            #  nor is there a next group.
            return squash_group_next()

        # ^^^

        def overlaps_current_fact(next_from_store):
            if next_from_store is None:
                return False
            return next_from_store.start < self.curr_group.time_until

        def fix_overlapping_fact(next_from_store):
            un_undoable_fact_next = next_from_store.copy()
            un_undoable_fact_next.start = self.curr_group.time_until
            un_undoable_fact_next.end = max(
                un_undoable_fact_next.end,
                self.curr_group.time_until,
            )
            un_undoable_fact_next.dirty_reasons.add('overlapped')
            # (lb): A hack to tell other UX components to alert user.
            un_undoable_fact_next.dirty_reasons.add('alert-user')
            return curr_group_add_next(un_undoable_fact_next)

        # ^^^

        def exists_in_between_groups(next_from_store):
            if next_from_store is None:
                return False
            next_group, next_group_index = fetch_next_group()
            if next_group is None:
                return True
            next_from_store_end = next_from_store.end or UntilTimeStops
            if next_from_store_end > next_group.time_since:
                return False
            self.controller.affirm(next_from_store.start >= self.curr_group.time_until)
            return True

        def fetch_next_group():
            next_group_index = self.groups.index(self.curr_group) + 1
            if next_group_index < len(self.groups):
                return self.groups[next_group_index], next_group_index
            return None, None

        # ^^^

        def squash_group_next():
            self.controller.affirm(self.curr_index == (len(self.curr_group) - 1))
            next_group, next_group_index = fetch_next_group()
            if next_group is None:
                return None
            return collapse_group(next_group, next_group_index)

        def collapse_group(next_group, next_group_index):
            next_fact = next_group[0]
            with self.fact_group_rekeyed():
                # The rekeyed fcn., pops curr_group, so decrement next index.
                next_group_index -= 1
                _next_group = self.groups.pop(next_group_index)
                self.controller.affirm(next_group is _next_group)
                self.curr_index += 1
                # Note that addition returns a new object, e.g.,
                #  self.curr_group += next_group
                # sets self.curr_group to a new object -- but fact_group_rekeyed
                # adds back the original group object, so do in-place addition.
                # Note that slice operator calls __setitem__ for each fact,
                # whereas addition calls __add__ or __radd__ just once.
                self.curr_group[:] = self.curr_group + next_group
            self.controller.affirm(self.curr_group[self.curr_index] is next_fact)
            return next_fact

        # ***

        def fill_gap_until(next_fact):
            if (next_fact is not None) and next_fact.has_prev_fact:
                self.controller.affirm(next_fact.prev_fact is self.curr_fact)
                return next_fact
            self.controller.affirm(not self.curr_fact.has_next_fact)
            if next_fact is None:
                gap_or_next = fill_gap_is_endless()
            else:
                gap_or_next = fill_gap_until_fact(next_fact)
            if gap_or_next is None:
                self.controller.affirm(False)  # 2020-04-15: (lb): Impossible?
                return None
            if gap_or_next is not next_fact:
                curr_group_add_next(gap_or_next)
            wire_two_facts_until(gap_or_next, next_fact)
            return gap_or_next

        def fill_gap_is_endless():
            gap_fact = self.fact_from_interval_gap(
                self.curr_fact.end, None,
            )
            return gap_fact

        def fill_gap_until_fact(next_fact):
            if next_fact.start == self.curr_fact.end:
                gap_fact = next_fact
            elif next_fact.is_gap:
                # Next fact is already unedited interval gap, so just edit its time.
                next_fact.start = self.curr_fact.end
                gap_fact = next_fact
            else:
                gap_fact = self.fact_from_interval_gap(
                    self.curr_fact.end, next_fact.start,
                )
            return gap_fact

        def wire_two_facts_until(gap_or_next, next_fact):
            if next_fact is not None:
                if next_fact is not gap_or_next:
                    self.wire_two_facts_neighborly(gap_or_next, next_fact)
                else:
                    self.wire_two_facts_neighborly(self.curr_fact, gap_or_next)
            self.controller.affirm(self.curr_fact.end == gap_or_next.start)
            self.controller.affirm(gap_or_next.prev_fact is self.curr_fact)
            self.controller.affirm(self.curr_fact.next_fact is gap_or_next)
            return gap_or_next

        return _jump_fact_inc()

    # ***

    def fetch_next_from_store(self, ref_time=None):
        ref_time = ref_time or self.curr_group.time_until
        # To process momentaneous Facts properly (and, e.g., to avoid finding
        # the same empty-time Fact again and again), send the Fact to the data
        # store lookup method, so that it can use the PK in the query.
        ref_fact = None
        if ref_time == self.curr_group[-1].end:
            ref_fact = self.curr_group[-1]
        # Search forward from the end time of the group (rather than,
        # say, calling subsequent(self.curr_fact)), so that we skip time
        # that's currently under our control.
        next_from_store = self.controller.facts.subsequent(
            fact=ref_fact, ref_time=ref_time,
        )
        if not next_from_store:
            # No more later facts from store.
            return None
        # Next fact should at least have a start time.
        self.controller.affirm(next_from_store.start)
        # In case we decide to keep this fact, make it safe.
        next_from_store.orig_fact = 0  # The orig_fact is... itself!
        # The caller will figure out if the next fact from the
        # store is one that is not yet under our control; or if it's
        # one we already know about (in which case, there exists a
        # group already blocking that time window).
        return next_from_store




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Defines dob command key mappings."""

import json

from gettext import gettext as _

from nark.config import ConfigRoot

__all__ = (
    'DobViewerConfigEditorKeys',
)


# ***

@ConfigRoot.section('editor-keys')
class DobViewerConfigEditorKeys(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # *** interface_keys.Key_Bonder.widget_focus()

    @property
    @ConfigRoot.setting(
        _("Switch to Next Widget (description → start time → end time → [repeats])"),
    )
    def focus_next(self):
        return 'tab'

    # ***

    @property
    @ConfigRoot.setting(
        _("Switch to Previous Widget (description → end time → start time → [repeats])"),
    )
    def focus_previous(self):
        return 's-tab'

    # ***

    @property
    @ConfigRoot.setting(
        _("Toggle To/From Start Time Widget"),
    )
    def edit_time_start(self):
        return 's'

    # ***

    @property
    @ConfigRoot.setting(
        _("Toggle To/From End Time Widget"),
    )
    def edit_time_end(self):
        return 'e'

    # *** interface_keys.Key_Bonder.save_and_quit()

    @property
    @ConfigRoot.setting(
        _("Save Changes"),
    )
    def save_edited_and_live(self):
        return 'c-s'

    # ***

    @property
    @ConfigRoot.setting(
        _("Save Changes and Exit"),
    )
    def save_edited_and_exit(self):
        # (lb): I had this mapped to Ctrl-W, but that feels like it should
        # just close the application... which would be a simple exit.
        # - I think Ctrl-W would break people's mental model:
        #     return 'c-w'
        # - But users can perform this same (two-commands-in-one) feature via
        #   the `:wq` commando. Except the commando has option to linger after
        #   save, to display the 'Saved {} Facts' message before exiting.
        # - Here we return the empty string to disable it, but it's left as a
        #   configurable option should the user want to map it to use it.
        return ''

    # ***

    @property
    @ConfigRoot.setting(
        _("Exit Quietly if No Changes"),
    )
    def exit_quietly(self):
        return 'q'

    # ***

    @property
    @ConfigRoot.setting(
        _("Exit with Prompt if Changes"),
    )
    def exit_command(self):
        # There are two Quit mapping: Ctrl-Q and ESCAPE.
        # - NOTE: Using 'escape' to exit is slow because PPT waits to
        #         see if escape sequence follows (which it wouldn't, after
        #         an 'escape', but meta-combinations start with an escape).
        #           tl;dr: 'escape' to exit is slow b/c alias resolution.
        # - NOTE: BUGBUG: Without the Ctrl-Q binding, if the user presses
        #         Ctrl-Q in the app., if becomes unresponsive.
        #         2020-04-11: I have not investigated, just noting it now
        #         that we're opening up keybindings for user to screw up! =)
        return 'c-q'

    # *** interface_keys.Key_Bonder.edit_time()

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_enter(self):
        return 'enter'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def toggle_focus_description(self):
        return 'd'

    # *** interface_keys.Key_Bonder.undo_redo()

    # Vim maps Ctrl-z and Ctrl-y for undo and redo;
    # and u/U to undo count/all and Ctrl-R to redo (count).
    #
    # (lb): So many undo/redo options!
    # MAYBE: Really mimic all of Vi's undo/redo mappings,
    #        or just pick one each and call it good?
    # - Or let user decide! 2020-04-11: Now it's customizable.

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def undo_command(self):
        return json.dumps([('c-z',), ('u',)])

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def redo_command(self):
        return json.dumps([('c-y',), ('c-r',), ('r',)])

    # *** interface_keys.Key_Bonder.normal()

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def rotate_help(self):
        return '?'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def dev_breakpoint(self):
        # I.e., 'm-=', aka, <Alt+=>.
        return json.dumps([('escape', '=')])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def jump_fact_dec(self):
        # I.e., 'j', or left arrow.
        return json.dumps([('j',), ('left',)])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def jump_fact_inc(self):
        # I.e., 'k', or right arrow.
        return json.dumps([('k',), ('right',)])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def jump_day_dec(self):
        # I.e., 'J', or Alt-left arrow.
        return json.dumps([('J',), ('escape', 'left')])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def jump_day_inc(self):
        # I.e., 'J', or Alt-right arrow.
        return json.dumps([('K',), ('escape', 'right')])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def jump_rift_dec(self):
        # A good pneumonic: 'F'irst Fact.
        return 'F'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def jump_rift_inc(self):
        # A good pneumonic: 'f'inal Fact.
        return 'f'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def jump_fact_first(self):
        return json.dumps([('g', 'g')])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def jump_fact_final(self):
        return 'G'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def cursor_up_one(self):
        return json.dumps([('h',), ('up',)])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def cursor_down_one(self):
        return json.dumps([('l',), ('down',)])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def scroll_up(self):
        return 'pageup'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def scroll_down(self):
        return 'pagedown'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def scroll_top(self):
        return 'home'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def scroll_bottom(self):
        return 'end'

    # *** interface_keys.Key_Bonder.edit_fact()

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_fact(self):
        # Edit merry-go-round: Prompt for act@gory, then tags, then description.
        return 'E'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_description(self):
        return 'd'

    # ***

    # Edit act@gory and tags using prompt__awesome.
    # (lb): This is pretty cool. prompt_awesome was built first,
    # and then I got comfortable with PPT and built the Carousel,
    # and then I was able to stick one inside the other, and it's
    # just awesome awesome now.

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_actegory(self):
        return 'a'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_tags(self):
        return 't'

    # *** interface_keys.Key_Bonder.nudge_time()

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_decrement_start(self):
        return json.dumps([('s-left',), (',',)])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_increment_start(self):
        return json.dumps([('s-right',), ('.',)])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_decrement_end(self):
        return json.dumps([('c-left',), ('[',)])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_increment_end(self):
        return json.dumps([('c-right',), (']',)])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_decrement_both(self):
        # FIXME/2019-01-21: Can you check if running in Terminator
        #  and warn-tell user. And/Or: Customize Key Binding feature.
        # In Terminator: Shift+Ctrl+Left/+Right: Resize the terminal left/right.
        #  (lb): I've disabled the 2 bindings in Terminator,
        #   so this works for me... so fixing it is a low priority!
        return 's-c-left'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_increment_both(self):
        # See previous comment about 's-c-left': The user's terminal probably
        # possibly has a mapping already that shadows this. (They can disable
        # their terminal mappings, though, for it to pass through.)
        return 's-c-right'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_decrement_start_5min(self):
        return '<'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_increment_start_5min(self):
        return '>'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_decrement_end_5min(self):
        return '{'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def edit_time_increment_end_5min(self):
        return '}'

    # *** interface_keys.Key_Bonder.create_delete_fact()

    # FIXME/2020-04-11: Not implemented.
    if False:

        @property
        @ConfigRoot.setting(
            _("XXX"),
        )
        def fact_split(self):
            return json.dumps([('escape', 'p')])

        # ***

        @property
        @ConfigRoot.setting(
            _("XXX"),
        )
        def fact_erase(self):
            return json.dumps([('escape', 'e')])

        # ***

        @property
        @ConfigRoot.setting(
            _("XXX"),
        )
        def fact_merge_prev(self):
            return json.dumps([('escape', 'm', 'left')])

        # ***

        @property
        @ConfigRoot.setting(
            _("XXX"),
        )
        def fact_merge_next(self):
            return json.dumps([('escape', 'm', 'right')])

    # *** interface_keys.Key_Bonder.clipboard()

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def fact_copy_fact(self):
        return 'c-c'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def fact_cut(self):
        return 'c-x'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def fact_paste(self):
        return 'c-v'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def fact_copy_activity(self):
        return json.dumps([('A', 'c-c')])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def fact_copy_tags(self):
        return json.dumps([('T', 'c-c')])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def fact_copy_description(self):
        return json.dumps([('D', 'c-c')])

    # *** interface_keys.Key_Bonder.vim_like_command*()

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def begin_commando(self):
        return ':'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def final_commando(self):
        return 'enter'

    # *** The ':' command strings.

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    # Should this be 'save_commando' for consistency with Ctrl-S command,
    # or should this be named with 'write' because 'w' and Vim suggest that?
    def write_commando(self):
        # As in, `:w`.
        return 'w'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def quit_commando(self):
        # As in, `:q`.
        return 'q'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def save_quit_commando(self):
        # As in, `:wq`.
        return 'wq'

    # *** interface_keys.Key_Bonder.delta_time_width()

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def begin_delta_time_start(self):
        return '+'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def begin_delta_time_end(self):
        return '-'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def allow_time_gap(self):
        return '!'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def final_delta_time_apply(self):
        return json.dumps([('enter',), ('tab',)])

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def final_delta_time_minutes(self):
        return 'm'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def final_delta_time_hours(self):
        return 'h'

    # ***

    # Reminder from nark's parse_time: iso8601 date formats:
    #   YYYY-MM-DD | YYYYMMDD | YYYY-MM | YYYY
    # Note that the only non-digit character allowed is the dash/minus.
    # - Here we allow the user to specify other characters that are okay
    #   to use as a separator, so we're not limiting to just dash.
    # - Note, too, that we can allow YYYYMM (sans dash) because why not.

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def date_separators(self):
        # NOTE: These are used in a regex, so escape as necessary.
        return json.dumps([
            # Conventional `YYYY-MM/DD` separators.
            ('-',), ('/',),
            # Convention separators between `YYYYMMDD` and `hh:mm`.
            (' ',), ('t',), ('T',),
            # Conventional `hh:mm` separator.
            (':',),
        ])

    # ***

    # Command shortcuts.

    # 2020-04-15: (lb): I did not consider much before picking the
    # key bindings below. There are not any obvious mappings, other
    # than perhaps (a) avoiding keys that have universal mappings
    # (like Ctrl-c, Ctrl-q, etc.). Also (b) making it easy to press
    # (i.e., not Ctrl-g), as these may be used often.
    # - I suppose between 'c-e' and 'V', on an English keyboard, they
    #   are similar motions (pinky plus pointer), so maybe that'll
    #   help people remember. And 'o' is pretty baked in from Vim.
    #   Open a new fact -- and prompt for the act@gory.

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def copy_complete_and_paste_active(self):
        return 'c-e'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def copy_complete_and_paste_new(self):
        return 'V'

    # ***

    @property
    @ConfigRoot.setting(
        _("XXX"),
    )
    def complete_and_prompt_new(self):
        # 'a' for act@gory, like plain 'a' command mapping.
        # Or not. 'c-a' is a common tmux prefix key.
        #  return 'c-a'
        # How 'bout, 'o'pen a new entry?
        return 'o'




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Fact-editing Redo/Undo Manager"""

import time
from collections import namedtuple

__all__ = (
    'RedoUndoEdit',
    'UndoRedoTuple',
)


UndoRedoTuple = namedtuple(
    'UndoRedoTuple', ('pristine', 'altered', 'time', 'what'),
)


class RedoUndoEdit(object):
    """"""
    def __init__(self, edits_manager):
        self.controller = edits_manager.controller
        self.debug = edits_manager.controller.client_logger.debug
        self.edits_manager = edits_manager
        self.undo = []
        self.redo = []

    # ***

    def append_changes(self, which, urt_changes, whence=''):
        which.append(urt_changes)
        # 2019-01-28: (lb): Added this method, and whence, to make
        # debugging easier for issues related to prev/next links.
        if self.controller.config['dev.catch_errors']:  # More devmode than catch_errors.
            facts_shorts = ''
            if urt_changes.altered:
                for idx, fact in enumerate(urt_changes.altered):
                    facts_shorts += ('\n- # {:d}.: {}'.format(idx, fact.short,))
            self.controller.client_logger.debug(
                '{}: no. changes: {} / to: {}{}'
                .format(
                    whence,
                    len(urt_changes.pristine),
                    which is self.undo and 'undo' or 'redo',
                    facts_shorts,
                ),
            )

    def clear_changes(self, which, whence=''):
        which.clear()
        if self.controller.config['dev.catch_errors']:
            self.controller.client_logger.debug(
                '{}: cleared changes from: {}'
                .format(
                    whence,
                    which is self.undo and 'undo' or 'redo',
                ),
            )

    # ***

    def add_undoable(self, copied_facts, what):
        undoable = UndoRedoTuple(copied_facts, None, time.time(), what)
        # Caller is responsible for calling update_undo_altered later.
        self.append_changes(self.undo, undoable, whence='add_undoable')

    def undo_peek(self):
        try:
            return self.undo[-1]
        except IndexError:
            return UndoRedoTuple([], [], None, None)

    # ***

    def undoable_changes(self, what, *edit_facts):
        edit_facts = list(filter(None, edit_facts))
        edit_fact_copies = [
            edit_fact.copy() for edit_fact in edit_facts if edit_fact is not None
        ]
        self.controller.affirm(len(edit_fact_copies) > 0)
        undoable_changes = UndoRedoTuple(
            edit_fact_copies, edit_facts, time.time(), what=what,
        )
        return undoable_changes

    # ***

    def remove_undo_if_nothing_changed(self, edit_facts):
        last_edits = self.undo_peek()
        if last_edits.pristine == edit_facts:
            # Nothing changed.
            toss_changes = self.undo.pop()
            self.debug('pop!: no.: {}'.format(len(toss_changes)))
            return None
        else:
            # Since what's on the undo is different, the redo is kaput.
            self.clear_changes(self.redo, 'remove_undo_if_nothing_changed')
            return last_edits.pristine

    # ***

    def update_undo_altered(self, edit_facts, append=False):
        try:
            undo_changes = self.undo.pop()
        except IndexError:
            self.controller.affirm(append)
            return

        self.controller.affirm(
            (
                (not append)
                and (
                    (undo_changes.altered is None)
                    or (undo_changes.altered == edit_facts)
                )
            )
            or (
                append
                and (
                    not set([fact.pk for fact in undo_changes.altered])
                    .intersection(set([fact.pk for fact in edit_facts]))
                )
            )
        )

        if append:
            edit_facts = undo_changes.altered + edit_facts

        undoable = UndoRedoTuple(
            undo_changes.pristine, edit_facts, undo_changes.time, undo_changes.what,
        )

        self.append_changes(
            self.undo,
            undoable,
            whence='update_undo_altered-{}'.format(undoable.what),
        )

        # This seals the deal; what's on the undo is different, and the redo is kaput.
        self.clear_changes(self.redo, 'update_undo_altered')

    # ***

    def undo_last_edit(self, restore_facts):
        try:
            undo_changes = self.undo.pop()
            self.debug('pop!: no.: {}'.format(len(undo_changes)))
        except IndexError:
            undone = False
        else:
            undone = True
            if restore_facts is not None:
                changes_copies = self.restore_facts(undo_changes, restore_facts)
                self.append_changes(self.redo, changes_copies, whence='undo_last_edit')
        return undone

    def redo_last_undo(self, restore_facts):
        try:
            redo_changes = self.redo.pop()
            self.debug('pop!: no.: {}'.format(len(redo_changes)))
        except IndexError:
            redone = False
        else:
            redone = True
            changes_copies = self.restore_facts(redo_changes, restore_facts)
            self.append_changes(self.undo, changes_copies, whence='redo_last_undo')
        return redone

    def restore_facts(self, fact_changes, restore_facts):
        restore_facts(fact_changes.pristine, fact_changes.altered)
        latest_changes = UndoRedoTuple(
            pristine=fact_changes.altered,
            altered=fact_changes.pristine,
            time=fact_changes.time,  # Not really meaningful anymore.
            what=fact_changes.what,
        )
        return latest_changes

    # ***

    # Combine edits into same undo if similar and made within short time
    # window, e.g, if user keeps adjusting time within 2-½ seconds of
    # previous adjustment, make just one undo object for whole operation.
    DISTINCT_CHANGES_THRESHOLD = 1.333

    def remove_undo_if_same_facts_edited(self, newest_changes):
        latest_changes = self.undo_peek()

        if latest_changes.what != newest_changes.what:
            self.debug('!what: no.: {}'.format(len(newest_changes.pristine)))
            return newest_changes

        if (
            (time.time() - latest_changes.time)
            > RedoUndoEdit.DISTINCT_CHANGES_THRESHOLD
        ):
            self.debug('!time: no.: {}'.format(len(newest_changes.pristine)))
            return newest_changes

        latest_pks = set([changed.pk for changed in latest_changes.pristine])
        if latest_pks != set([edit_fact.pk for edit_fact in newest_changes.pristine]):
            self.debug('!pks: no.: {}'.format(len(newest_changes.pristine)))
            return newest_changes

        latest_undo = self.undo.pop()
        self.debug('pop!: no.: {}'.format(len(latest_undo.pristine)))
        return latest_undo




"""
Sphinx Read the Docs theme.

From https://github.com/ryan-roemer/sphinx-bootstrap-theme.
"""

from os import path

import sphinx


__version__ = '0.4.3.dev0'
__version_full__ = __version__


def get_html_theme_path():
    """Return list of HTML theme paths."""
    cur_dir = path.abspath(path.dirname(path.dirname(__file__)))
    return cur_dir


# See http://www.sphinx-doc.org/en/stable/theming.html#distribute-your-theme-as-a-python-package
def setup(app):
    app.add_html_theme('sphinx_rtd_theme', path.abspath(path.dirname(__file__)))

    if sphinx.version_info >= (1, 8, 0):
        # Add Sphinx message catalog for newer versions of Sphinx
        # See http://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx.application.Sphinx.add_message_catalog
        rtd_locale_path = path.join(path.abspath(path.dirname(__file__)), 'locale')
        app.add_message_catalog('sphinx', rtd_locale_path)

    return {'parallel_read_safe': True, 'parallel_write_safe': True}



# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Zone Manager"""

from gettext import gettext as _

import os
from inflector import English, Inflector

import click_hotoffthehamster as click

# Profiling: load prompt_toolkit. ~ 0.040 secs.
from prompt_toolkit.application import Application
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import (
    FloatContainer,
    HorizontalAlign,
    HSplit,
    VSplit
)
from prompt_toolkit.output.color_depth import ColorDepth
from prompt_toolkit.styles import Style
from prompt_toolkit.widgets import Box, Label

from dob_bright.crud.facts_diff import FactsDiff

from ..ptkui.dialog_overlay import alert_and_question

from .exceptions import catch_action_exception
from .zone_content import ZoneContent
from .zone_details import ZoneDetails
from .zone_lowdown import ZoneLowdown
from .zone_streamer import ZoneStreamer

__all__ = (
    'ZoneManager',
)


class ZoneManager(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel
        self.facts_diff = None

        self.zone_streamer = ZoneStreamer(self.carousel)
        self.zone_details = ZoneDetails(self.carousel)
        self.zone_content = ZoneContent(self.carousel)
        self.zone_lowdown = ZoneLowdown(self.carousel)

        self.alert_showing = False
        self.silence_alert_overlapped = False

    # ***

    def standup(self):
        self.zone_streamer.standup()
        self.zone_details.standup()
        self.zone_content.standup()
        self.zone_lowdown.standup()
        self.assemble_focus_jumps()

    # ***

    def build_and_show(self, **kwargs):
        self.root = self.build_root_container()
        self.layout = self.build_application_layout()
        self.setup_styling()
        self.center_thyself()
        self.application = self.build_application_object(**kwargs)
        self.rebuild_viewable()

    # ***

    def build_root_container(self):
        self.streamer_posit = 0
        self.details_posit = 1
        self.content_posit = 2
        self.lowdown_posit = 3
        self.hsplit = HSplit(
            # Use placeholders; we'll populate in rebuild_containers().
            children=[
                Label(text=''),  # zone_streamer
                Label(text=''),  # zone_details
                Label(text=''),  # zone_content
                Label(text=''),  # zone_lowdown
            ],
            # There can be top-bottom padding, e.g.,:
            #  padding=3,
            #  padding_char='X',
            #  padding_style='',
        )

        if self.carousel.style_classes['editor-align'] == 'LEFT':
            app_align = HorizontalAlign.LEFT
            app_width = click.get_terminal_size()[0]
        elif self.carousel.style_classes['editor-align'] == 'CENTER':
            app_align = HorizontalAlign.CENTER
            app_width = None
        elif self.carousel.style_classes['editor-align'] == 'RIGHT':
            app_align = HorizontalAlign.RIGHT
            app_width = click.get_terminal_size()[0]
        else:
            # self.carousel.style_classes['editor-align'] == 'JUSTIFY'
            app_align = HorizontalAlign.JUSTIFY
            app_width = None

        self.vsplit = VSplit(
            [self.hsplit],
            align=app_align,
            width=app_width,
        )

        root_container = FloatContainer(Box(body=self.vsplit), floats=[])
        return root_container

    def build_application_layout(self):
        layout = Layout(
            container=self.root,
            # Will get set later:
            #   focused_element
            # EXPLAIN/2019-01-21: How does focused_element get set?
            #  Automatically? Side effect of another call?
            #  The Layout() constructor has a focused_element attr,
            #   but we do not need to set it.
        )
        return layout

    def setup_styling(self):
        class_styles = self.carousel.style_classes['collect_tups']
        try:
            self.style = Style(class_styles)
        except Exception as err:
            # FIXME: Show error in Carousel, or raise and show before first draw.
            #        - You could show error on startup, pause for input,
            #        then continue (could even decide not to show warning again?).
            # FIXME: Find all controller usage from clyde and refactor so has to
            #        be passed in (wire what you need from dob upfront).
            from dob_bright.termio import dob_in_user_warning
            msg = _('The user style “{0}” failed to load: {1}').format(
                self.carousel.controller.config['editor.styling'], str(err),
            )
            self.carousel.controller.client_logger.warning(msg)
            dob_in_user_warning(str(class_styles))
            dob_in_user_warning(msg)
            self.style = Style([])

    def center_thyself(self):
        if not self.carousel.controller.config['editor.centered']:
            return
        click.clear()
        # (lb): Revisit this? A little hacky.
        # Newlines seem to nudge Carousel centered.
        # Not sure why 3 is magic number, or if always.
        click.echo()
        click.echo()
        click.echo()

    def _detect_color_depth(self):
        # MAYBE/2020-01-06: Make color_depth configurable. Or detect better.
        # (lb): This is a little frustrating. Colors are true in mate-terminal, but
        #   in tmux, they're being "rounded", e.g., my custom class:category-sleep,
        #   which is #CA85AC pink, is being rounded to straight up 0xFF0000 red.
        #   - I have TERM=xterm in both cases (albeit not advisable in tmux, it's
        #     the only TERM I can get italics to work in). And when I trace code
        #     here, it seems like TERM defaults to DEPTH_8_BIT in either case --
        #     but for some reason, under tmux, there's color rounding.
        #   - One solution/work-around is to use ColorDepth.DEPTH_24_BIT when...
        #     unfortunately, I'm not sure the best way to detect if the terminal
        #     supports truecolor. Obvi., for me, I'll always have a truecolor
        #     terminal at my fingers, or will fix it if I don't but for other
        #     users, I want to ensure they have a decent experience always.
        #     - We can check COLORTERM, which I think is set by mate-terminal,
        #       except when I searched the source, I did not see it therein.
        #       But works for me! (I mean, we could just always set 24-bit color,
        #       but this at least falls back to not doing anything if COLORTERM is
        #       not set as expected, which seems like the safest course of action.)
        # Default, which works in raw mate-terminal, but not under tmux:
        #   color_depth=ColorDepth.DEPTH_8_BIT,
        if os.environ.get('COLORTERM', None) != 'truecolor':
            return None
        return ColorDepth.DEPTH_24_BIT

    def build_application_object(self, **kwargs):
        # (lb): By default, the app uses editing_mode=EditingMode.EMACS,
        # which adds a few key bindings. One binding in particular is a
        # little annoying -- ('c-x', 'c-x') -- because PPT has to wait
        # for a second key press, or a timeout, to resolve the binding.
        # E.g., if you press 'c-x', it takes a sec. until our handler is
        # called (or, it's called if you press another key, but then the
        # response seems weird, i.e., 2 key presses are handled seemingly
        # simultaneously after the second keypress, rather than being
        # handled individually as the user presses them keys). In any
        # case -- long comment! -- set editing_mode to something other
        # than EditingMode.EMACS or EditingMode.VI (both are just strings).
        # FIXME/2019-11-23: (lb): That comment is old, because we do not
        # disable editing_mode, but we leave it set to emacs for some
        # other features. Question is, is there a delay on Ctrl-x? And
        # what does Ctrl-X do? Is it meta-cut? Also, what are the emacs-
        # binding features that you like?
        application = Application(
            layout=self.layout,
            key_bindings=self.carousel.action_manager.key_bindings_normal,
            full_screen=False,
            color_depth=self._detect_color_depth(),
            erase_when_done=True,
            # Enables mouse wheel scrolling.
            # CAVEAT: Steals from terminal app!
            #   E.g., while app is active, mouse wheel scrolling no
            #   longer scrolls the desktop Terminal app window, ha!
            # FIXME: Make this feature optionable. Seems like some
            #   people may appreciate this wiring.
            mouse_support=True,
            # (lb): I tried changing style at runtime, e.g., editing a class
            # name's 'fg:#xxxxxx' value in the Application style_rules, but
            # it did not have an impact. So stick to adding and removing
            # class names from each widget's style component, which works.
            style=self.style,
            # The scrollable content area disables input with
            #   read_only=True
            # which we could do here by disabling built-in buffer
            # editing bindings, i.e.,
            #   editing_mode='',
            # but we want those bindings for the command inputizer.
            **kwargs,
        )
        return application

    # ***

    def rebuild_viewable(self):
        """
        rebuild_viewable is called to update the view after the user edits a
        fact or navigates to a different fact.

        HINT: If the view gets messed up, say by a pdb session, this function
        will not redraw everything. See instead:

            self.application.renderer.clear()
        """
        self.reset_diff_fact()
        self.rebuild_containers()
        self.selectively_refresh()
        self.carousel.controller.client_logger.debug(_('rebuilt and refreshed'))

    def reset_diff_fact(self):
        orig_fact = self.carousel.edits_manager.curr_orig
        # Call editable_fact, which either gets the edit_fact, or gets
        # a copy of the orig_fact; but it does not make an undo.
        edit_fact = self.carousel.edits_manager.editable_fact()
        self.facts_diff = FactsDiff(orig_fact, edit_fact, formatted=True)
        self.carousel.controller.client_logger.debug(
            'facts_diff: {}'.format(self.facts_diff),
        )

    def rebuild_containers(self):
        streamer_container = self.zone_streamer.rebuild_viewable()
        self.hsplit.get_children()[self.streamer_posit] = streamer_container

        details_container = self.zone_details.rebuild_viewable()
        self.hsplit.get_children()[self.details_posit] = details_container

        content_container = self.zone_content.rebuild_viewable()
        self.hsplit.get_children()[self.content_posit] = content_container

        lowdown_container = self.zone_lowdown.rebuild_viewable()
        self.hsplit.get_children()[self.lowdown_posit] = lowdown_container

    # ***

    def selectively_refresh(self):
        self.zone_streamer.selectively_refresh()
        self.zone_details.selectively_refresh()
        self.zone_content.selectively_refresh()
        self.zone_lowdown.selectively_refresh()

    # ***

    def assemble_focus_jumps(self):
        content_control = self.content_control
        start_control = self.widget_control_time_start
        end_control = self.widget_control_time_end

        self.focus_chain = [
            content_control,
            start_control,
            end_control,
        ]

        self.focus_surround = {
            content_control: self.zone_content.focus_content,
            start_control: self.zone_details.edit_time_start,
            end_control: self.zone_details.edit_time_end,
        }

        self.focus_recent = content_control

    @catch_action_exception
    def focus_next(self, event):
        # Note also:
        #   prompt_toolkit.key_binding.bindings.focus.focus_next
        self.focus_move(lambda index: (index + 1) % len(self.focus_chain))

    @catch_action_exception
    def focus_previous(self, event):
        # Note also:
        #   prompt_toolkit.key_binding.bindings.focus.focus_previous
        self.focus_move(lambda index: (index or len(self.focus_chain)) - 1)

    def focus_move(self, index_f):
        # NOTE: PTK has focus_next(event), which can cycle through focusable
        #       containers, but the edit-time controls are on-demand, so we
        #       do this manually.
        curr_control = self.layout.current_control
        try:
            defocused = self.focus_surround[curr_control](focus=False)
        except Exception as err:  # noqa: F841 local variable ... never used
            # 2020-04-07: See d4d12cc3. This should be fixed!
            self.carousel.controller.affirm(False)  # REPL if dev.catch_errors.
            return
        if not defocused:
            return
        curr_index = self.focus_index(curr_control)
        next_index = index_f(curr_index)
        next_control = self.focus_chain[next_index]
        self.focus_surround[next_control](focus=True)
        self.focus_recent = curr_control

    def focus_index(self, which_control=None):
        if which_control is None:
            which_control = self.layout.current_control
        which_index = self.focus_chain.index(which_control)
        return which_index

    @property
    def content_control(self):
        return self.zone_content.content.control

    @property
    def widget_control_time_end(self):
        return self.zone_details.widgets_end.text_area.control

    @property
    def widget_control_time_start(self):
        return self.zone_details.widgets_start.text_area.control

    # ***

    @catch_action_exception
    # SKIP: @ZoneContent.Decorators.reset_showing_help
    def toggle_focus_time_end(self, event):
        self.toggle_focus_time_widget(self.widget_control_time_end)

    @catch_action_exception
    # SKIP: @ZoneContent.Decorators.reset_showing_help
    def toggle_focus_time_start(self, event):
        self.toggle_focus_time_widget(self.widget_control_time_start)

    def toggle_focus_time_widget(self, time_control):
        focus = self.layout.current_control is not time_control
        if focus:
            which_index = self.focus_index(time_control)
        else:
            # (lb): Toggling to recent sorta feels weird.
            #   which_index = self.focus_index(self.focus_recent)
            # Toggling between content and time feels more natural.
            which_index = self.focus_index(self.content_control)
        self.focus_move(lambda index: which_index)

    def ppt_control_buffer_avoid_selection(self, control, mouse_event):
        # python-prompt-toolkit/prompt_toolkit/layout/controls.py's
        # BufferControl.mouse_handler
        # on MouseEventType.MOUSE_UP
        # calls buffer.start_selection
        # if user clicks anywhere except the cursor_position,
        # so ensure the cursor_position is set on MOUSE_DOWN
        # to the click, otherwise the MOUSE_UP that is about
        # to happen will create an awkward selection in the
        # widget.
        buffer = control.buffer
        index = buffer.document.translate_row_col_to_index(
            mouse_event.position.y, mouse_event.position.x,
        )
        buffer.exit_selection()
        buffer.cursor_position = index

    # ***

    def jump_msg_with_count(self, count, direction, what):
        return '{} {} {}'.format(
            direction,
            count,
            Inflector(English).conditional_plural(count, what),
        )

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def jump_fact_dec(self, event):
        """"""
        count = self.carousel.update_handler.apply_count_multiplier()
        jump_msg = self.jump_msg_with_count(count, _('Backward'), _('Fact'))
        prev_fact = self.carousel.edits_manager.jump_fact_dec(count=count)
        self.finalize_jump_dec(prev_fact, jump_msg)

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def jump_fact_inc(self, event):
        """"""
        count = self.carousel.update_handler.apply_count_multiplier()
        jump_msg = self.jump_msg_with_count(count, _('Forward'), _('Fact'))
        next_fact = self.carousel.edits_manager.jump_fact_inc(count=count)
        self.finalize_jump_inc(next_fact, jump_msg)

    # ***

    # (lb): NOTE/2019-01-24: There are no explicit jump-week or jump-month,
    # etc., commands, but if the user keeps the 'J' or 'K' keys pressed
    # (the prev and next day commands) for an extended length of time, the
    # Carousel will start jumping by larger time increments.

    def jump_msg_count_and_time(self, count, direction, what):
        return '{} ({} {})'.format(
            self.jump_msg_with_count(count, direction, what),
            _('from'),
            str(self.carousel.edits_manager.conjoined.jump_time_reference),
        )

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def jump_day_dec(self, event):
        """"""
        count = self.carousel.update_handler.apply_count_multiplier(floats=True)
        jump_msg = self.jump_msg_count_and_time(count, _('Backward'), _('Day'))
        prev_fact = self.carousel.edits_manager.jump_day_dec(days=count)
        self.finalize_jump_dec(prev_fact, jump_msg)

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def jump_day_inc(self, event):
        """"""
        count = self.carousel.update_handler.apply_count_multiplier(floats=True)
        jump_msg = self.jump_msg_count_and_time(count, _('Forward'), _('Day'))
        next_fact = self.carousel.edits_manager.jump_day_inc(days=count)
        self.finalize_jump_inc(next_fact, jump_msg)

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def jump_rift_dec(self, event):
        """"""
        rift_jumper = self.carousel.edits_manager.jump_rift_dec
        noop_msg = _("Already on first Fact")
        self.jump_rift_or_time(rift_jumper, 'until_time', noop_msg)

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def jump_rift_inc(self, event):
        """"""
        rift_jumper = self.carousel.edits_manager.jump_rift_inc
        noop_msg = _("Already on final Fact")
        self.jump_rift_or_time(rift_jumper, 'since_time', noop_msg)

    def jump_rift_or_time(self, rift_jumper, which_time, noop_msg):
        was_curr = self.carousel.edits_manager.curr_fact
        modifier = self.carousel.update_handler.command_modifier
        self.carousel.update_handler.command_modifier_reset()
        if not modifier:
            rift_jumper()
        else:
            jump_time, parse_err = self.zone_details.parse_dated(modifier)
            if parse_err is not None:
                noop_msg = _("Not a time or date")
            else:
                facts_mgr = self.carousel.edits_manager.conjoined
                kwargs = {which_time: jump_time}
                facts_mgr.jump_to_fact_nearest(**kwargs)
                noop_msg = _("Nothing on that date")
        self.refresh_fact_or_notify_noop(was_curr, noop_msg)

    def refresh_fact_or_notify_noop(self, was_curr, noop_msg):
        if was_curr is not self.carousel.edits_manager.curr_fact:
            self.rebuild_viewable()
        else:
            self.update_status(noop_msg)

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def jump_fact_first(self, event):
        """"""
        was_curr = self.carousel.edits_manager.curr_fact
        self.carousel.edits_manager.jump_fact_first()
        self.refresh_fact_or_notify_noop(was_curr, _("Already on first Fact"))

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def jump_fact_final(self, event):
        """"""
        was_curr = self.carousel.edits_manager.curr_fact
        self.carousel.edits_manager.jump_fact_final()
        self.refresh_fact_or_notify_noop(was_curr, _("Already on final Fact"))

    # ***

    def finalize_jump_dec(self, prev_fact, jump_msg):
        """"""
        self.finalize_jump_check_overlapped(
            prev_fact, noop_msg=_("Viewing earliest Fact"), jump_msg=jump_msg,
        )

    def finalize_jump_inc(self, next_fact, jump_msg):
        """"""
        self.finalize_jump_check_overlapped(
            next_fact, noop_msg=_("Viewing latest Fact"), jump_msg=jump_msg,
        )

    def finalize_jump_check_overlapped(self, curr_fact, noop_msg, jump_msg=''):
        def _finalize_jump_check_overlapped():
            _jump_msg = jump_msg
            if curr_fact and 'alert-user' in curr_fact.dirty_reasons:
                curr_fact.dirty_reasons.discard('alert-user')
                # 2019-02-13: (lb): Currently, only 'overlapped' causes this.
                self.carousel.controller.affirm(
                    curr_fact.dirty_reasons == set(['overlapped']),
                )
                popop_modal_alert_overlapped_fact()
                _jump_msg = _('ALERT! Corrected overlapping Fact times. Save to commit.')
            self.finalize_jump(curr_fact, noop_msg, _jump_msg)

        def popop_modal_alert_overlapped_fact():
            if self.silence_alert_overlapped:
                return
            alert_and_question(
                self.root,
                title=_('Overlapping Fact'),
                label_text=_(
                    'A Fact loaded from the data store overlaps an adjacent Fact.'
                ) + '\n\n' + _(
                    'The Fact has been updated and is staged to be saved.'
                ),
                prompt_ok=_('Got it!'),
                prompt_no=_('Keep reminding me'),
                on_close=self.on_alert_overlapped_close,
            )
            # Disable any input recognize (let PPT's dialog handle everything).
            self.carousel.action_manager.wire_keys_modal()

        _finalize_jump_check_overlapped()

    def on_alert_overlapped_close(self, result):
        self.alert_showing = False
        if result:
            self.silence_alert_overlapped = True
        # Re-enable keyboard input processing.
        self.carousel.action_manager.wire_keys_normal()

    def finalize_jump(self, curr_fact, noop_msg, jump_msg=''):
        if curr_fact is None:
            self.update_status(noop_msg)
        else:
            self.zone_content.reset_cursor_left_column()
            self.rebuild_viewable()
            if jump_msg:
                self.update_status(jump_msg)

    # ***

    def update_status(self, hot_notif):
        self.zone_lowdown.update_status(hot_notif)




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""GroupChained, an ordered Facts list."""

from gettext import gettext as _

from nark.items.fact import SinceTimeBegan, UntilTimeStops
from sortedcontainers import SortedKeyList

__all__ = (
    'GroupChained',
    'sorted_facts_list',
)


# ***

def integer_range_groupify(integers):
    from itertools import groupby
    from operator import itemgetter

    def _integer_range_groupify():
        srtd = sorted(map(int, integers))
        return assemble_groups(srtd)

    def range_key_recursive(srtd, ix):
        if ix[0] == 0:
            # Very first item, so always start of group.
            return ix[1]
        prev_idx = ix[0] - 1
        prev_val = int(srtd[prev_idx])
        if ix[1] == prev_val + 1:
            # Previous item value is 1 before this one, so walk backwards.
            return range_key_recursive(srtd, (prev_idx, prev_val,))
        return ix[1]

    def range_key(srtd, ix):
        key = range_key_recursive(srtd, ix)
        # print('ix:', ix, 'key:', key)
        return key

    def assemble_groups(srtd):
        grouped = []
        for key, grp in groupby(enumerate(srtd), lambda ix: range_key(srtd, ix)):
            # print('key: {} / grp: {} {}'.format(key, grp, list(grp)))
            # print(list(map(itemgetter(1), grp)))
            grouped.append(list(map(itemgetter(1), grp)))
        return grouped

    return _integer_range_groupify()


# ***

def sorted_facts_list(facts=None):
    sorted_facts_list = SortedKeyList(
        iterable=facts,
        # Use sorty_times, not sorty_tuple, to ignore PKs. Otherwise, two facts
        # with the same PK would not appear as equivalent, which would make the
        # fact-finding insertion/update/removal code less easily maintainable.
        key=lambda item: item.sorty_times,
    )
    return sorted_facts_list


# ***

class GroupChained(object):
    """
    A GroupChained represents an ordered list of Facts.
    In the context of the Carousel, the ordered list
    is also guaranteed to be contiguous, and may contain
    a mix of stored Facts, new Facts, and gap Facts.
    Or even no Facts, if group is simply claiming time.
    """
    # FIXME/2019-12-06: (lb): Just testing. Remove affirm arg. later.
    # def __init__(self, facts=None):
    def __init__(self, facts=None, affirm=None):
        self.facts = sorted_facts_list(facts)
        self.reset_time_window()
        self.affirm = affirm or (lambda _x: None)

    # ***

    def __add__(self, other):
        return GroupChained(facts=self.facts + other.facts)

    def __radd__(self, other):
        return GroupChained(facts=other.facts + self.facts)

    # ***

    def __delitem__(self, key):
        del self.facts[key]

    def __getitem__(self, key):
        return self.facts[key]

    # For, e.g., self.group[:] = ...
    def __setitem__(self, key, value):
        del self.facts[key]
        try:
            # For, e.g., self.group[:] = ...
            # value is a slice().
            for fact in value.facts:
                self.facts.add(fact)
            self.claim_fact_time(value.facts)
        except AttributeError:
            # For, e.g., self.group[0] = ...
            # value is a (Placeable)Fact.
            self.facts.add(value)

    # ***

    def __iter__(self):
        return iter(self.facts)

    def __len__(self):
        return len(self.facts)

    # ***

    def __eq__(self, other):
        if self is other:
            return True
        if (other is not None) and isinstance(other, GroupChained):
            other = other.times
        return self.times == other

    def __gt__(self, other):
        return self.times > other.times

    def __lt__(self, other):
        return self.times < other.times

    # ***

    def __str__(self):
        def range_str(grp):
            if len(grp) == 0:
                return 'No IDs!'
            elif len(grp) == 1:
                return '{}'.format(grp[0])
            else:
                lhs = grp[0]
                rhs = grp[-1]
                if grp[1] < 0:
                    lhs, rhs = rhs, lhs
                return '{} to {}'.format(lhs, rhs)

        def assemble_pk_ranges():
            # 2019-12-03: (lb): Cull Fact(s) with pk == None.
            # - I fixed a bug: I added a tag, and saved. But self.facts here
            # had a fact with pk == None, to wit: str(None) returned 'None',
            # and integer_range_groupify blew up on ValueError.
            # - I'd like to confidently note why facts.pk is None, so affirming.
            facts = list(filter(None, self.facts))
            # FIXME/2019-12-06: (lb): Just testing. Remove affirm later.
            self.affirm(len(facts) == len(self.facts))
            pks = [str(fact.pk) for fact in facts]
            grouped = integer_range_groupify(pks)
            pk_ranges = ', '.join([range_str(grp) for grp in grouped])
            return len(pks), pk_ranges

        num_pks, pk_ranges = assemble_pk_ranges()
        return _(
            "‘{0}’ to ‘{1}’ / No. Facts: {2} / PK(s): {3}"
        ).format(self.time_since, self.time_until, num_pks, pk_ranges)

    # ***

    def claim_fact_time(self, facts=None):
        """NOTE: This function changes the SortedKeyList key, beware!"""
        if facts is None:
            facts = self.facts
        if not facts:
            return
        until = facts[-1].end or UntilTimeStops
        self.claim_time_span(since=facts[0].start, until=until)

    def claim_time_span(self, since=None, until=None):
        """NOTE: This function changes the SortedKeyList key, beware!"""
        # Handle since time.
        if self.time_since is None:
            self.time_since = since
        elif since is not None:
            self.time_since = min(self.time_since, since)
        # Handle until time.
        if self.time_until is None:
            self.time_until = until
        elif until is not None:
            self.time_until = max(self.time_until, until)

    def contains_fact_time(self, facts=None):
        if facts is None:
            facts = self.facts
        if not facts:
            return
        if (
            (facts[0].start < self.time_since)
            or (
                (facts[-1].end is None)
                and (self.time_until < UntilTimeStops)
            )
            or (
                (facts[-1].end is not None)
                and (facts[-1].end > self.time_until)
            )
        ):
            return False
        return True

    @property
    def sorty_times(self):
        return (self.time_since, self.time_until, )

    def reset_time_window(self):
        # Default to time window not defined (as opposed to empty
        # (impossible) time window achieved with (UntilTimeStops,
        # SinceTimeBegan)).
        self.time_since = None
        self.time_until = None
        self.claim_fact_time()

    @property
    def since_time_began(self):
        return self.time_since is SinceTimeBegan

    @property
    def until_time_stops(self):
        return self.time_until is UntilTimeStops

    # ***

    def add(self, some_fact):
        # Caller Beware: This changes the group key!
        self.facts.add(some_fact)
        self.claim_fact_time([some_fact])

    def bisect_key_left(self, key):
        return self.facts.bisect_key_left(key)

    def bisect_left(self, value):
        return self.facts.bisect_left(value)

    def index(self, some_fact):
        for index, fact in enumerate(self.facts):
            if some_fact.pk == fact.pk:
                return index
        raise ValueError("Fact with PK '{0}' is not in list".format(some_fact.pk))

    def pop(self, index):
        return self.facts.pop(index)




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""dob_viewer.config sub.package provides Carousel UX user configuration settings."""

import json

from gettext import gettext as _

__all__ = (
    'json_load_sublisted',
)


def json_load_sublisted(cfgname, cfgval):
    def _json_load_sublisted():
        # (lb): We could skip the startswith check and just use except,
        #       but it feels more readable this way.
        if not cfgval.startswith('['):
            # Just a string. Except don't bother with the empty string,
            # which is used to disable a key command mapping.
            if cfgval:
                return [cfgval], None
            return [], None
        try:
            # List of lists: Top-level is list, and elements are lists.
            # - This is currently used for keybindings, which can either
            #   be a single element sublist (representing a single key),
            #   or the sublist could be a key binding tuple (well, it's
            #   json, so a list), which represents a multiple-key binding.
            keycodes = json.loads(cfgval)
            assert isinstance(keycodes, list)  # Would it be anything else?
            if not sanity_check(keycodes):
                return None, error_not_list_within_lists()
        except json.decoder.JSONDecodeError as err:
            return None, error_not_list_within_lists(err)
        return keycodes, None

    def sanity_check(keycodes):
        return all(isinstance(keycode, list) for keycode in keycodes)

    def error_not_list_within_lists(err=''):
        append_err = ' (“{}”)'.format(err) if err else ''
        return (_(
            'ERROR: Key binding for ‘{}’ should be single key'
            ' or list of lists, not: {}{}'
            .format(cfgname, cfgval, append_err)
        ))

    return _json_load_sublisted()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""``hamster``, ``hamster``, ``hamster``!!! a cuddly, furry time tracker."""

import os
import sys

from nark import get_version as _get_version

__all__ = (
    'get_version',
    '__arg0name__',
    '__package_name__',
)

# Note that this package is a library, so __arg0name__ likely, e.g., 'dob'.
__arg0name__ = os.path.basename(sys.argv[0])

__package_name__ = 'dob-viewer'


def get_version(include_head=False):
    return _get_version(
        package_name=__package_name__,
        reference_file=__file__,
        include_head=include_head,
    )




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

""""""

__all__ = (
    'FactsManager_Rift',
)


class FactsManager_Rift(object):
    """"""
    def __init__(self, *args, **kwargs):
        super(FactsManager_Rift, self).__init__()

        self.time_rifts = []

    # ***

    def place_time_rifts(self):
        def _place_time_rifts():
            self.time_rifts = []
            for group in self.groups:
                first_fact = self.find_first_dirty_from_group(group)
                if first_fact is None:
                    first_fact = group[0]
                add_time_rift(first_fact)
            last_group_last_fact = self.groups[-1][-1]
            if self.time_rifts[-1] != last_group_last_fact.start:
                add_time_rift(last_group_last_fact)

        def add_time_rift(some_fact):
            self.controller.client_logger.debug(
                'time_rifts: {}'.format(some_fact.start),
            )
            self.time_rifts.append(some_fact.start)

        _place_time_rifts()

    def find_rift_fact(self, is_next=False, is_prev=False):
        def _find_rift_fact():
            assert is_next ^ is_prev
            prev_time, next_time = drop_time_bounds()
            since_time = next_time if is_next else None
            until_time = prev_time if is_prev else None
            if not since_time and not until_time:
                return None
            rift_fact = self.jump_to_fact_nearest(
                since_time=since_time,
                until_time=until_time,
            )
            return rift_fact

        def drop_time_bounds():
            prev_start = None
            for start_time in self.time_rifts:
                if self.curr_fact.start < start_time:
                    return (prev_start, start_time)
                if start_time != self.curr_fact.start:
                    prev_start = start_time
            return (prev_start, None)

        return _find_rift_fact()

    def find_first_dirty(self):
        for group in self.groups:
            first_dirty_fact = self.find_first_dirty_from_group(group)
            if first_dirty_fact is not None:
                return first_dirty_fact
        # None edited/dirty. Fall back on last fact (most recent).
        return self.groups[-1][-1]

    def find_first_dirty_from_group(self, group):
        for fact in group:
            if fact.dirty:
                return fact
        return None




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""FactsManager_FactDec"""

import datetime

from nark.items.fact import SinceTimeBegan

__all__ = (
    'FactsManager_FactDec',
)


JULIAN_YEAR = datetime.timedelta(days=365, hours=6)  # 365.25 days, 31557600 secs.


# SYNC_ME: Keep FactsManager_FactDec and FactsManager_FactInc synced.
#   That, or write a generic base class, but that sounds painfuller to maintain.
#   then simply running meld as necessary:
#
#       meld facts_mgr_fact_dec.py facts_mgr_fact_inc.py &
class FactsManager_FactDec(object):
    """"""
    def __init__(self, controller, *args, **kwargs):
        super(FactsManager_FactDec, self).__init__()
        # FIXME/2018-07-31: Make user_start_time (beginning_of_time) configurable.
        #   E.g., user could set their birthday (birthtime).
        #   Could also maybe not use default if feature not enabled,
        #     but for now, using 100 yrs. ago.
        self.beginning_of_time = controller.now - (100 * JULIAN_YEAR)

    def jump_fact_dec(self):
        """"""
        def _jump_fact_dec():
            # Check first if we've reached the beginning of time.
            is_first_fact = (self.curr_index == 0)
            if is_first_fact and self.curr_group.since_time_began:
                return None
            # Each facts group represents a contiguous block of time that is
            # being managed by dob, and to ignore anything in the store from
            # that time. Find either the previous fact under dob's control,
            # or one from the store (that we'll wire into dob in a later step).
            prev_fact = fetch_prev_fact()
            # Each fact's prev_fact and next_fact links are not wired until
            # we inspect them here for gaps -- meaning, we fill in the gap facts
            # reactively, as the user traverses the facts in the carousel. Now.
            prev_fact = fill_gap_since(prev_fact)
            # If we've reached the beginning of time, there's nothing previous.
            if prev_fact is None:
                return None
            # Just a sanity check. Note that we use fact.times for >, otherwise
            # sorty_tuple is used, which also sorts by PK, which masks 2 facts
            # at same time but with unique IDs.
            # - This also shows that we update self.curr_index already, and
            #   maybe self.curr_group, but not self.curr_fact (so the state
            #   is outta sorts).
            self.controller.affirm(self.curr_fact.start >= prev_fact.end)
            # We (re)wired the fact to the group earlier; now rewire the group.
            self.fulfill_jump(prev_fact, reason='fact-dec')
            # See if we've identified the boundary of the known factiverse.
            if prev_fact.start <= self.beginning_of_time:
                with self.fact_group_rekeyed():
                    self.curr_group.claim_time_span(since=SinceTimeBegan)
                self.controller.affirm(self.curr_group.since_time_began)
            self.controller.client_logger.debug('\n- prev: {}'.format(prev_fact.short))
            return prev_fact

        # ***

        def fetch_prev_fact():
            # Check curr_fact.curr_index, and not curr_fact.has_prev_fact,
            # because the facts may not all be wired yet.
            if self.curr_index > 0:
                return from_curr_group_decrement()
            else:
                # Choose between fact from store or previous group (if any).
                return from_prev_group_or_store()

        # ***

        def from_curr_group_decrement():
            self.curr_index -= 1
            prev_from_curr_group = self.curr_group[self.curr_index]
            return prev_from_curr_group

        # ***

        def curr_group_add_prev(prev_fact):
            self.controller.affirm(self.curr_fact.prev_fact is None)
            self.curr_group_add(prev_fact)
            if self.curr_group[self.curr_index] is not prev_fact:
                # Added gap-fact, which comes after previous fact
                # we already decremented to reference.
                self.curr_index += 1
            self.controller.affirm(self.curr_group[self.curr_index] is prev_fact)
            return prev_fact

        # ***

        def from_prev_group_or_store():
            prev_from_store = self.fetch_prev_from_store()
            if overlaps_current_fact(prev_from_store):
                return fix_overlapping_fact(prev_from_store)
            if exists_in_between_groups(prev_from_store):
                # Add the fact to the group, but do not wire yet,
                # as we may add a gap fact, if necessary.
                return curr_group_add_prev(prev_from_store)
            # No facts between this group and previous group;
            # or there are no facts from store
            #  nor is there a previous group.
            return squash_group_prev()

        # ^^^

        def overlaps_current_fact(prev_from_store):
            if prev_from_store is None:
                return False
            return prev_from_store.end > self.curr_group.time_since

        def fix_overlapping_fact(prev_from_store):
            un_undoable_fact_prev = prev_from_store.copy()
            un_undoable_fact_prev.end = self.curr_group.time_since
            un_undoable_fact_prev.start = min(
                un_undoable_fact_prev.start,
                self.curr_group.time_since,
            )
            un_undoable_fact_prev.dirty_reasons.add('overlapped')
            # (lb): A hack to tell other UX components to alert user.
            un_undoable_fact_prev.dirty_reasons.add('alert-user')
            return curr_group_add_prev(un_undoable_fact_prev)

        # ^^^

        def exists_in_between_groups(prev_from_store):
            if prev_from_store is None:
                return False
            prev_group, prev_group_index = fetch_prev_group()
            if prev_group is None:
                return True
            if prev_from_store.start < prev_group.time_until:
                return False
            self.controller.affirm(prev_from_store.end <= self.curr_group.time_since)
            return True

        def fetch_prev_group():
            prev_group_index = self.groups.index(self.curr_group) - 1
            if prev_group_index >= 0:
                return self.groups[prev_group_index], prev_group_index
            return None, None

        # ^^^

        def squash_group_prev():
            self.controller.affirm(self.curr_index == 0)
            prev_group, prev_group_index = fetch_prev_group()
            if prev_group is None:
                return None
            return collapse_group(prev_group, prev_group_index)

        def collapse_group(prev_group, prev_group_index):
            prev_fact = prev_group[-1]
            with self.fact_group_rekeyed():
                _prev_group = self.groups.pop(prev_group_index)
                self.controller.affirm(prev_group is _prev_group)
                self.curr_index = len(prev_group) - 1
                # Note that addition returns a new object, e.g.,
                #  self.curr_group = prev_group + self.curr_group
                # sets self.curr_group to a new object -- but fact_group_rekeyed
                # adds back the original group object, so do in-place addition.
                # Note that slice operator calls __setitem__ for each fact,
                # whereas addition calls __add__ or __radd__ just once.
                self.curr_group[:] = prev_group + self.curr_group
            self.controller.affirm(self.curr_group[self.curr_index] is prev_fact)
            return prev_fact

        # ***

        def fill_gap_since(prev_fact):
            if (prev_fact is not None) and prev_fact.has_next_fact:
                self.controller.affirm(prev_fact.next_fact is self.curr_fact)
                return prev_fact
            self.controller.affirm(not self.curr_fact.has_prev_fact)
            if prev_fact is None:
                gap_or_prev = fill_gap_since_users_life_began()
            else:
                gap_or_prev = fill_gap_since_prev_fact(prev_fact)
            if gap_or_prev is None:
                return None
            if gap_or_prev is not prev_fact:
                curr_group_add_prev(gap_or_prev)
            wire_two_facts_since(gap_or_prev, prev_fact)
            return gap_or_prev

        def fill_gap_since_users_life_began():
            gap_fact = None
            user_start_time = self.beginning_of_time
            if self.curr_fact.start > user_start_time:
                gap_fact = self.fact_from_interval_gap(
                    user_start_time, self.curr_fact.start,
                )
            return gap_fact

        def fill_gap_since_prev_fact(prev_fact):
            if prev_fact.end == self.curr_fact.start:
                gap_fact = prev_fact
            elif prev_fact.is_gap:
                # Prior fact is already unedited interval gap, so just edit its time.
                prev_fact.end = self.curr_fact.start
                gap_fact = prev_fact
            else:
                gap_fact = self.fact_from_interval_gap(
                    prev_fact.end, self.curr_fact.start,
                )
            return gap_fact

        def wire_two_facts_since(gap_or_prev, prev_fact):
            if prev_fact is not None:
                if prev_fact is not gap_or_prev:
                    self.wire_two_facts_neighborly(prev_fact, gap_or_prev)
                else:
                    self.wire_two_facts_neighborly(gap_or_prev, self.curr_fact)
            self.controller.affirm(gap_or_prev.end == self.curr_fact.start)
            self.controller.affirm(gap_or_prev.next_fact is self.curr_fact)
            self.controller.affirm(self.curr_fact.prev_fact is gap_or_prev)
            return gap_or_prev

        return _jump_fact_dec()

    # ***

    def fetch_prev_from_store(self, ref_time=None):
        ref_time = ref_time or self.curr_group.time_since
        # Momentaneous Facts support.
        ref_fact = None
        if ref_time == self.curr_group[0].start:
            ref_fact = self.curr_group[0]
        # Search backward from the start time of the group (rather than,
        # say, calling antecedent(self.curr_fact)), so that we skip time
        # that's currently under our control.
        prev_from_store = self.controller.facts.antecedent(
            fact=ref_fact, ref_time=ref_time,
        )
        if not prev_from_store:
            # No more prior facts from store.
            return None
        # Prior fact should at least have a start time.
        self.controller.affirm(prev_from_store.start)
        # In case we decide to keep this fact, make it safe.
        prev_from_store.orig_fact = 0  # The orig_fact is... itself!
        # The caller will figure out if the previous fact from the
        # store is one that is not yet under our control; or if it's
        # one we already know about (in which case, there exists a
        # group already blocking that time window).
        return prev_from_store




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel Header (Fact meta and diff)"""

from gettext import gettext as _

from datetime import datetime
import re

from prompt_toolkit.application.current import get_app
from prompt_toolkit.layout.containers import HSplit, VSplit, to_container
from prompt_toolkit.lexers import SimpleLexer
from prompt_toolkit.widgets import Label, TextArea

from nark.helpers.parse_errors import ParserInvalidDatetimeException
from nark.helpers.parse_time import parse_dated

from dob_bright.crud.fix_times import must_complete_times

from ..ptkui.dialog_overlay import show_message

from .exceptions import catch_action_exception
from .zone_details_time_end import ZoneDetails_TimeEnd
from .zone_details_time_start import ZoneDetails_TimeStart

__all__ = (
    'ZoneDetails',
)


class ZoneDetails(
    ZoneDetails_TimeStart,
    ZoneDetails_TimeEnd,
):
    """"""
    def __init__(self, carousel):
        super(ZoneDetails, self).__init__()
        self.carousel = carousel
        self.active_widgets = None
        # Convenience attrs.
        self.affirm = self.carousel.controller.affirm
        self.debug = self.carousel.controller.client_logger.debug

    class HeaderKeyVal(object):
        """"""
        def __init__(
            self,
            index,
            what_part=None,
            fact_attr=None,
            diff_kwargs=None,
            key_parts=None,
            val_label=None,
            text_area=None,
            orig_val=None,
            mouse_handler=None,
        ):
            self.index = index
            self.what_part = what_part
            self.fact_attr = fact_attr
            self.diff_kwargs = diff_kwargs
            self.key_parts = key_parts
            self.val_label = val_label
            self.text_area = text_area
            self.orig_val = orig_val
            self.mouse_handler = mouse_handler

    # ***

    def standup(self):
        """"""
        def _standup():
            # A couple convenience attrs.
            self.zone_manager = self.carousel.zone_manager
            self.zone_content = self.carousel.zone_manager.zone_content
            assemble_children()
            self.details_container = build_container()

        def assemble_children():
            self.children = []
            add_meta_lines()

        def add_meta_lines():
            # Skipping: add_header_midpoint.
            add_header_duration()
            self.add_header_start_time()
            self.add_header_end_time()
            # Skipping: add_header_fact_pk.
            # Skipping: add_header_deleted.
            # Skipping: add_header_split_from.
            add_header_activity()
            add_header_category()
            add_header_tags()
            add_blank_line()

        # ***

        def add_header_duration():
            # MEH/2019-11-22: (lb): Duration label mouse handler, but to do what?
            # - User clicks, modal asks for new duration, adjust end time to match?
            self.label_duration = self.add_header_section('duration')

        def add_header_activity():
            self.widgets_activity = self.add_header_section(
                'activity',
                'activity_name',
                mouse_handler=header_widget_mouse_handler('actegory'),
            )

        def add_header_category():
            self.widgets_category = self.add_header_section(
                'category',
                'category_name',
                mouse_handler=header_widget_mouse_handler('actegory'),
            )

        def add_header_tags():
            # FIXME/BACKLOG/2018-07-19: Long tags can extend width,
            #   but do not wrap the line. Need to manually break?
            #   See: wrap_on_whitespace_maybe
            # FIXME/2018-07-28: Fix tags display: with 2+ tags, inserting
            #   newlines makes the height dance. But keeping for now, as not
            #   many Facts (o' mine) with two or more tags.
            self.widgets_tags = self.add_header_section(
                'tags',
                'tags_tuples',
                split_lines=True,
                colorful=True,
                mouse_handler=header_widget_mouse_handler('tags'),
            )

        def add_blank_line():
            self.blank_line = self.make_section_component('', style='class:blank-line')
            # Note that PTK add a default style, class:label, before our style, e.g.,,
            #    self.blank_line.window.style == 'class:label class:blank-line'
            # Which we'll keep, because it makes setting an application background easy.
            self.children.append(self.blank_line)

        # ***

        def build_container():
            details_container = HSplit(children=self.children)
            return details_container

        # ***

        def header_widget_mouse_handler(restrict_edit):
            def _mouse_handler(mouse_event):
                # MAYBE/2019-11-22: (lb): Interesting feature, but jarring:
                # - User could click on Activity name in header, or Category,
                # or Tags, and app could respond by showing Awesome Prompt.
                # However, this behavior is rather jarring if you're not
                # expecting it.
                # - Also, user might be trying to click and drag to select
                # the Activity or Category name, but immediately on mouse
                # down the Awesome Prompt is shown (this callback is called),
                # and PPT doesn't seem to bother sending us more than a single
                # mouse down. So unless PPT updated to not trigger callback
                # if user either selecting text or double clicking, and to
                # only trigger callback on simple, non-drag press and release,
                # this feature should only be enabled if user wants it on.
                # - MAYBE/2019-11-22: (lb): Maybe add config option to enable.
                #   Until then, disabled by way of hidden branching.
                if False:
                    self.carousel.enduring_edit = True
                    self.carousel.restrict_edit = restrict_edit
                    get_app().exit()
            return _mouse_handler

        # ***

        _standup()

    # ***

    def rebuild_viewable(self):
        """"""
        self.refresh_all_children()
        return self.details_container

    # ***

    def add_header_section(
        self,
        part_type,
        fact_attr=None,
        editable=False,
        mouse_handler=None,
        **kwargs
    ):
        text_area = None
        if editable:
            # 2019-11-22: (lb): Some PPT controls have focus options, e.g.,
            #   TextArea(..., focusable=True, focus_on_click=True)
            # but the Carousel handles changing focus, so we instead
            # register a mouse handler to change the focus ourselves.
            # (diff_attrs() wires the handler using the third element
            # of the ubiquitous (style, text) tuples). Note also that
            # I tried options, focusable=True and focus_on_click=True,
            # and nothing.
            # (lb): I tried making a custom lexer, e.g., class MyLexer(Lexer),
            # but for whatever reason, the `assert isinstance(lexer, Lexer)`
            # in PPT's DynamicLexer fails, because the custom lexer object
            # somehow reports lexer.__class__: abc.ABCMeta. Fortunately,
            # rather than trying to explain how the lexer object gets
            # converted into some meta wrapper thing, we can use a builtin
            # lexer that simply styles the input text.
            text_area = TextArea(
                height=1,
                lexer=SimpleLexer(style='class:value-focus'),
            )

        keyval_parts = ZoneDetails.HeaderKeyVal(
            index=len(self.children),
            what_part=part_type,
            fact_attr=fact_attr,
            diff_kwargs=kwargs,
            key_parts=self.make_header_label_parts(part_type),
            val_label=self.make_header_value_part(part_type),
            text_area=text_area,
            mouse_handler=mouse_handler,
        )
        self.children.append(
            VSplit(
                children=[
                    *keyval_parts.key_parts,
                    keyval_parts.val_label,
                ],
            )
        )
        return keyval_parts

    def make_header_label_parts(self, part_type=''):
        name = _(part_type)
        prefix = '  '
        padded = '{:.<19}'.format(name)
        kv_sep = ' : '

        tline_style, title_style = self.header_line_styles(part_type, set_focus=False)

        # Note that Label() inserts its own style, 'class:label', before
        # the styles we pass it. We'll leave it, because it lets user put
        # `label = 'fg#<>'` in their styles.conf to (almost) achieve a
        # uniform application background (save for the content area).
        labels = [
            self.make_section_component(
                prefix, style=tline_style, dont_extend_width=True,
            ),
            self.make_section_component(
                padded, style=title_style, dont_extend_width=True,
            ),
            self.make_section_component(
                kv_sep, style=tline_style, dont_extend_width=True,
            ),
        ]
        return labels

    def make_header_value_part(self, part_type=''):
        style = 'class:value-normal-line '
        if part_type:
            style += 'class:value-{}-line '.format(part_type)
        return self.make_section_component('', style=style)

    # ***

    def make_section_component(
        self, header_text='', style='', dont_extend_width=False,
    ):
        # The header label is called once when first showing the Fact,
        # and not called during the rebuild_viewable heartbeat. It's
        # not even rebuilt when focus changes between components.
        label = Label(
            text=header_text,
            style=style,
            dont_extend_width=dont_extend_width,
        )
        return label

    # ***

    def refresh_all_children(self):
        self.refresh_duration()
        self.refresh_time_start()
        self.refresh_time_end()
        self.refresh_activity()
        self.refresh_category()
        self.refresh_tags()
        self.refresh_blank_line()

    # ***

    def selectively_refresh(self):
        # Update times and spans based off <now>.
        self.refresh_duration()
        # Update start time, should its time have been adjusted.
        self.refresh_time_start()
        # Update the <now> time duration that FactsDiff shows.
        self.refresh_time_end()

    # ***

    def refresh_duration(self):
        # The style_class is 'class:value-normal class:value-duration '.
        style_class = self.assemble_style_class_for_part(self.label_duration)
        orig_val, edit_val = self.zone_manager.facts_diff.diff_time_elapsed(
            show_now=True, style_class=style_class,
        )
        diff_tuples = self.zone_manager.facts_diff.diff_line_tuples_style(
            orig_val, edit_val, style_class=style_class,
        )
        self.label_duration.val_label.text = diff_tuples
        self.process_style_rules(self.label_duration)

    def refresh_activity(self):
        self.refresh_val_widgets(self.widgets_activity)

    def refresh_category(self):
        self.refresh_val_widgets(self.widgets_category)

    def refresh_tags(self):
        self.refresh_val_widgets(self.widgets_tags)

    def refresh_blank_line(self):
        # Lets the user override the blank line style for matching rules.
        custom_classes = self.carousel.process_style_rules(
            ppt_widget=None, friendly_name='blank-line',
        )
        # - (lb): Rules replace, not append, widget's style. #rule_replace
        #   I.e., not;
        #       ...style = 'class:label class:blank-line ' + custom_classes
        #   This was an arbitrary choice. If it's better the other way, we can switch.
        self.blank_line.window.style = custom_classes or 'class:label class:blank-line '

    def refresh_val_widgets(self, keyval_widgets):
        self.affirm(keyval_widgets.fact_attr)
        # The style_class is 'class:value-normal class:value-{activity|category|etc} '.
        style_class = self.assemble_style_class_for_part(keyval_widgets)
        diff_tuples = self.zone_manager.facts_diff.diff_attrs(
            keyval_widgets.fact_attr,
            style_class=style_class,
            mouse_handler=keyval_widgets.mouse_handler,
            **keyval_widgets.diff_kwargs
        )
        keyval_widgets.val_label.text = diff_tuples
        # We've already set the default value (on top of PPT's Label default class):
        #   keyval_widgets.val_label.window.style:
        #     'class:label class:value-normal-line'
        # and now we'll set value-{normal|activity|category|etc}[-line], if rules apply.
        # (lb): Note also widgets_start and widgets_end come through here.
        self.process_style_rules(keyval_widgets)

    def assemble_style_class_for_part(self, keyval_widgets):
        style_class = 'class:value-normal '
        style_class += 'class:value-{} '.format(keyval_widgets.what_part)
        return style_class

    def process_style_rules(self, keyval_parts, set_focus=None):
        # (lb): Reminder that rules replace, not append, widget's style. #rule_replace
        self.process_style_rules_header_titles(keyval_parts, set_focus)
        self.process_style_rules_header_values_normal(keyval_parts)
        self.process_style_rules_header_values_focus(keyval_parts, set_focus)

    def process_style_rules_header_titles(self, keyval_parts, set_focus=None):
        # keyval_parts.key_parts is a list of Labels (from make_header_label_parts).
        # - The first item is the left column padding. E.g.: '  '.
        # - The second item is the meta label and the ...-padding.
        # - The third item is the middle column, ' : ', before the value.
        # (lb): This is another 'MEH' moment: How much customizability
        #   do we really need? Or, put another way, we can always add
        #   more later. For now, it seems like enough to be able to both
        #   customize the inner meta label (including the .-padding), or
        #   to customize the whole meta line. I don't see a need (or, I
        #   don't have a need) to customize the left or center column bits.
        #   (Also, you can customize specially when focused, so there are
        #   already four ways to customize each header title.)
        if set_focus is None and self.active_widgets is keyval_parts:
            set_focus = True
        dot_padded_title = keyval_parts.key_parts[1]
        title_part = 'title-{}'.format(keyval_parts.what_part)
        if not set_focus:
            friendlies = ['title-normal', title_part]
        else:
            # This keyval_parts should be editable.
            self.affirm(keyval_parts.text_area is not None)
            friendlies = ['title-focus', title_part + '-focus']

        for friendly_name in friendlies:
            # Register the -line style, which means assigns same style to all parts.
            for label in keyval_parts.key_parts:
                self.carousel.process_style_rules(label, friendly_name + '-line')
            # Register title-[normal|duration|start|end|activity|category|tags][-focus]
            # on just the title label part of the header.
            self.carousel.process_style_rules(dot_padded_title, friendly_name)

    def process_style_rules_header_values_normal(self, keyval_parts):
        # For matching rules, apply rule styles for: value-normal, value-normal-line,
        # and value-[duration|start|end|activity|category|tags][-line]
        value_part = 'value-{}'.format(keyval_parts.what_part)
        for friendly_name in ('value-normal', value_part):
            for suffix in ('-line', ''):
                self.carousel.process_style_rules(
                    keyval_parts.val_label,
                    friendly_name + suffix,
                )

    def process_style_rules_header_values_focus(self, keyval_parts, set_focus=False):
        if not set_focus:
            return

        # For matching rules, apply any rule styles for:
        # value-focus, value-focus-line, and value-[start|end]-focus[-line].
        value_part = 'value-{}-focus'.format(keyval_parts.what_part)
        for friendly_name in ('value-focus', value_part):
            for suffix in ('-line', ''):
                self.carousel.process_style_rules(
                    keyval_parts.text_area,
                    friendly_name + suffix,
                )

    # ***

    def header_line_styles(self, part_type, set_focus=False):
        focus_state = set_focus and 'focus' or 'normal'
        # Set classes on each part, title-normal-line, or title-focus-line.
        # Set also on inner text, title-normal or title-focus.
        title_base_class = 'class:title-{}'.format(focus_state)
        title_line_class = '{}-line'.format(title_base_class)
        # Set classes more specific to the part type, e.g.,
        # title-[duration|start|end|activity|category|tags][-focus][-line].
        tpart_base_class = 'class:title-{}'.format(part_type)
        if set_focus:
            tpart_base_class += '-{}'.format(focus_state)  # += '-focus'
        tpart_line_class = '{}-line'.format(tpart_base_class)

        line_classes = '{} {} '.format(title_line_class, tpart_line_class)
        text_classes = '{} {} {} {} '.format(
            title_line_class, title_base_class, tpart_line_class, tpart_base_class,
        )

        return line_classes, text_classes

    # ***

    def replace_val_container(
        self,
        val_container,
        keyval_widgets,
        set_focus=False,
    ):
        keyval_vsplit = self.details_container.get_children()[keyval_widgets.index]

        focus_state = set_focus and 'focus' or 'normal'

        line_classes, text_classes = self.header_line_styles(
            keyval_widgets.what_part, set_focus=set_focus,
        )
        # (lb): PTK assigns the same 'class:label' to all widgets. If we maintain it,
        # then the user can set, e.g., "label = 'bg:#00AA66'" in their styles.conf to
        # set a universal background color (well, except for content area and border).
        ptk_label_class = 'class:label '
        line_classes = ptk_label_class + line_classes
        text_classes = ptk_label_class + text_classes

        keyval_vsplit.get_children()[0].style = line_classes

        title_widget = keyval_vsplit.get_children()[1]
        title_widget.style = text_classes

        keyval_vsplit.get_children()[2].style = line_classes

        # ***

        # On Label, to_container gets val_container.formatted_text_control.
        # On TextArea, to_container gets val_container.window.
        value_widget = to_container(val_container)
        # Set class on value component, value-normal-line, or value-focus-line.
        value_line_class = 'class:value-{}-line '.format(focus_state)
        # Set class on value component, value-{part}[-focus]-line,
        # e.g., value-end-focus-line, or value-start-line.
        value_part_class = 'class:value-{}'.format(keyval_widgets.what_part)
        if set_focus:
            value_part_class += '-focus'
        value_part_class += '-line'
        # This clobbers default style that PPT use on TextArea, 'class:text-area '.
        value_widget.style = ptk_label_class + value_line_class + value_part_class
        keyval_vsplit.get_children()[3] = value_widget

        # ***

        self.process_style_rules(keyval_widgets, set_focus=set_focus)

    def replace_val_container_label(self, keyval_widgets):
        self.replace_val_container(
            keyval_widgets.val_label,
            keyval_widgets,
        )
        # keyval_widgets.val_label.window.style: 'class:value-normal-line'
        #   and formatted_text_control.style: ''
        #   and keyval_widgets.val_label.text
        #     is style tuples, [('class:value-normal', '...')...]
        keyval_widgets.val_label

    def replace_val_container_text_area(self, keyval_widgets):
        self.replace_val_container(
            keyval_widgets.text_area,
            keyval_widgets,
            set_focus=True,
        )

    # ***

    def edit_time_focus(self, keyval_widgets):
        self.active_widgets = keyval_widgets
        # Swap out a container in the layout.
        self.replace_val_container_text_area(keyval_widgets)
        # Focus the newly placed container.
        # MAYBE/2019-01-28: (lb): This is highly coupled, to say the least.
        #   Should maybe have zone_manager pass a focus callback.
        self.carousel.zone_manager.layout.focus(keyval_widgets.text_area)
        # Move the cursor to the end of the exit field,
        # e.g., if there's a date and time already set,
        # put the cursor after it all.
        self.send_cursor_right_to_end(keyval_widgets.text_area.buffer)
        # Wire a few simple bindings for editing (mostly rely on PPT's VI mode.)
        self.carousel.action_manager.wire_keys_edit_time()

    # ***

    def edit_time_leave(self, keyval_widgets):
        def _edit_time_leave():
            self.affirm(
                (self.active_widgets is None)
                or (keyval_widgets is self.active_widgets)
            )
            return apply_edited_and_refresh()

        def apply_edited_and_refresh():
            leave_okayed = not was_edited()
            if not leave_okayed:
                leave_okayed = self.edit_time_enter(passive=True)
            if not leave_okayed:
                return False
            return refresh_keyval()

        def was_edited():
            debug_log_text()
            return keyval_widgets.text_area.text != keyval_widgets.orig_val

        def debug_log_text():
            self.debug('text_area.text: {} / orig_val: {}'.format(
                keyval_widgets.text_area.text,
                keyval_widgets.orig_val,
            ))

        def refresh_keyval():
            # Refresh labels now, so that old value isn't shown briefly and then
            # updated, which looks weird. Rather, update label first, then show.
            self.selectively_refresh()
            self.replace_val_container_label(self.active_widgets)
            self.active_widgets = None
            return True

        return _edit_time_leave()

    # ***

    def send_cursor_right_to_end(self, winbufr):
        end_posit = winbufr.document.get_end_of_document_position()
        # Generally same as: winbufr.document.get_end_of_line_position()
        winbufr.cursor_right(end_posit)

    # ***

    def edit_time_any_key(self, event=None):
        self.debug('event: {}'.format(event))
        # Ignore all alpha characters except those for [t|T]imezone delimiter.
        if event.data.isalpha() and event.data not in ('t', 'T'):
            return
        # Like PPT's basic binding's filter=insert_mode, or vi's filter=vi_replace_mode.
        # "Insert data at cursor position."
        # PPT basic binding's self-insert:
        #   event.current_buffer.insert_text(event.data * event.arg)
        # PPT vi binding's vi_replace_mode:
        #  event.current_buffer.insert_text(event.data, overwrite=True)
        event.current_buffer.insert_text(event.data)
        self.editable_was_edited = True

    # ***

    # FIXME/BACKLOG/2019-01-21: Notify user if time changed to not delete adjacent.
    #   - Test current behavior. Old hamster would delete facts if you extended one
    #     and showed others. I don't like that behavior! User can delete Facts first,
    #     and then fill in the empty time! (Or someone else can submit a PR so long
    #     as the destructive behavior is not the default.)
    #   - Probably use footer to show message; but could instead use popup modal.

    # FIXME/BACKLOG/2019-01-21: Need way to cancel after editing time:
    #   Possible key binding: Ctrl-q, ESC, q, etc.

    @catch_action_exception
    def edit_time_enter(self, event=None, passive=False):
        """"""
        leave_okayed = [True, ]

        def _edit_time_enter():
            edit_text = self.active_widgets.text_area.text
            # Note that carousel.edits_manager.curr_edit returns fact-under-edit
            # only if one already exists, but fact may be unedited, in which case
            # it'd return the original, unedited fact. So use the editable fact we
            # made earlier.
            edit_fact = self.zone_manager.facts_diff.edit_fact
            apply_edited_time(edit_fact, edit_text)
            return leave_okayed[0]

        def apply_edited_time(edit_fact, edit_text):
            self.debug('edit_time_enter: edit_text: {}'.format(edit_text))
            if not edit_text:
                apply_edit_time_removed(edit_fact)
            else:
                apply_edit_time_changed(edit_fact, edit_text)

        # ***

        def apply_edit_time_removed(edit_fact):
            if self.active_widgets is self.widgets_start:
                okay = self.apply_edit_time_removed_start(edit_fact, passive)
                if not okay:
                    # User hit 'enter'. Annoy them with a warning.
                    show_message_cannot_clear_start()
                # else, passive=True, and widget was reset to orig start
                # (because removing a Fact's start time is never allowed).
            else:
                self.affirm(self.active_widgets is self.widgets_end)
                okay = self.apply_edit_time_removed_end(edit_fact)
                if okay:
                    apply_edit_time_valid(edit_fact, edit_time=None)
                else:
                    # Always warn user, whether they hit 'enter' or are tabbing away.
                    show_message_cannot_clear_end()

        # ***

        def apply_edit_time_changed(edit_fact, edit_text):
            edit_time, parse_err = self.parse_dated(edit_text)
            if parse_err:
                show_message_cannot_parse_time(parse_err)
            else:
                apply_edit_time_valid(edit_fact, edit_time)

        def apply_edit_time_valid(edit_fact, edit_time):
            was_fact = edit_fact.copy()
            if self.active_widgets is self.widgets_start:
                was_time = edit_fact.start_fmt_local
                applied = self.apply_edit_time_start(
                    edit_fact, edit_time, verify_fact_times,
                )
            else:
                self.affirm(self.active_widgets is self.widgets_end)
                was_time = edit_fact.end_fmt_local_or_now
                applied = self.apply_edit_time_end(
                    edit_fact, edit_time, verify_fact_times,
                )
            check_conflicts_and_confirm(edit_fact, was_fact, was_time, applied)

        def verify_fact_times(edit_fact):
            # Check that edit_fact.start < edit_fact.end now, rather than
            # via apply_edit_time_start/apply_edit_time_end's call to
            # edits_manager.apply_edits. By doing it now, we will not have
            # to cleanup the undo stack, nor fix the current fact, nor any
            # neighbor fact.
            conflicts = must_complete_times(
                self.carousel.controller,
                new_facts=[edit_fact],
                progress=None,
                leave_blanks=True,
                other_edits={},
                suppress_barf=True,
            )
            if conflicts:
                edited_fact_alert_conflicts(conflicts)
                return False
            return True

        def check_conflicts_and_confirm(edit_fact, was_fact, was_time, applied):
            if not applied:
                # Nothing changed; no-op.
                return
            edited_fact_check_conflicts(edit_fact, was_fact, was_time)

        def edited_fact_check_conflicts(edit_fact, was_fact, was_time):
            # (lb): The application interface mostly precludes conflicts.
            #   The apply_edit_time_start/apply_edit_time_end methods do
            #   not allow conflicts with other Facts. But user can create a
            #   conflict with the Fact itself, e.g., by changing the end time
            #   to come before the start.
            conflicts = edited_fact_conflicts(edit_fact)
            if conflicts:
                edited_fact_alert_conflicts(conflicts)
                edited_fact_cleanup_rejected(edit_fact, was_fact, was_time)

        def edited_fact_alert_conflicts(conflicts):
            self.affirm(len(conflicts) == 1)
            conflict_msg = conflicts[0][2]
            show_message_conflicts_found(conflict_msg)

        def edited_fact_cleanup_rejected(edit_fact, was_fact, was_time):
            # Toss the rejected edit from the undo stack.
            undone = self.carousel.edits_manager.toss_last_edit()
            self.affirm(undone)
            # Reset the Fact, otherwise the edit will stick!
            #   # Won't work! Screws up prev/next links, and leaves undo!:
            #   edit_fact.start = was_fact.start
            #   edit_fact.end = was_fact.end
            # MAYBE/2020-04-09: (lb): Pretty sure we're missing code to
            # reset the fact. (This code path used to be taken when a fact
            # end time < its start time. But that case is now handled earlier,
            # and the flow won't make it here. And if you set a fact's start
            # time earlier than the previous fact start time, it'll be auto-
            # corrected. So I'm not quite sure what use case triggers this
            # branch anymore!)
            self.affirm(edit_fact.start == was_fact.start)
            self.affirm(edit_fact.end == was_fact.end)
            # FIXME/2020-04-09: (lb): Seriously, what gets you here?
            # - Document what use case gets you here if/when this fires.
            self.affirm(False)
            # Update the control text.
            self.active_widgets.text_area.text = was_time

        def edited_fact_conflicts(edit_fact):
            # Skip all unstored, edited Facts.
            other_edits = {
                fact.pk: fact for fact in self.carousel.edits_manager.prepared_facts
            }

            conflicts = must_complete_times(
                self.carousel.controller,
                new_facts=[edit_fact],
                progress=None,
                leave_blanks=True,
                other_edits=other_edits,
                suppress_barf=True,
            )
            self.debug('no. conflicts found: {}'.format(len(conflicts)))
            return conflicts

        # ***

        def show_message_cannot_clear_start():
            show_message_and_deny_leave(
                self.carousel.zone_manager.root,
                _('Try again'),
                _(
                    "You may not clear a Fact's start time.\n\n"
                    "Enter a valid date and time, clock time, or relative time."
                ),
            )

        def show_message_cannot_clear_end():
            show_message_and_deny_leave(
                self.carousel.zone_manager.root,
                _('You lose'),
                _("You may not clear a Fact's end time unless it is the final Fact."),
            )

        def show_message_cannot_parse_time(edit_text):
            show_message_and_deny_leave(
                self.carousel.zone_manager.root,
                _('Drat!'),
                edit_text,
            )

        def show_message_conflicts_found(conflict_msg):
            show_message_and_deny_leave(
                self.carousel.zone_manager.root,
                _('Not so fast!'),
                conflict_msg,
            )

        def show_message_and_deny_leave(*args, **kwargs):
            leave_okayed[0] = False
            show_message(*args, **kwargs)

        # ***

        def re_focus_maybe():
            if passive:
                # User is tabbing; caller is handling focus.
                return
            self.zone_content.focus_content()

        # ***

        return _edit_time_enter()

    # ***

    def parse_dated(self, edit_text):
        def _parse_dated():
            time_now = self.carousel.controller.now
            parseable = prepare_dated()
            try:
                edit_time = parse_dated(parseable, time_now, cruftless=True)
            except ParserInvalidDatetimeException as err:
                # E.g., try entering date "2019-01-27 18."
                edit_time = None
                parse_err = str(err)
            else:
                self.affirm(isinstance(edit_time, datetime))
                parse_err = None
            return edit_time, parse_err

        def prepare_dated():
            # Per date_separators, we let user specify which optional
            # separators they'd like to, but iso8601 only allows dashes to
            # separate the year parts; a space, 't', or 'T' to distinguish
            # the year from the time; and colons to separate clock time's
            # hours from minutes. Here we remove all those non-numeric
            # characters and just build the iso8601-worthy string ourselves.
            # We also allow incomplete YYYYMM (iso8601 only allows YYYY-MM
            # or YYYYMMDD), as well as incomplete time (say, '12', for noon).
            numberless = re.sub(r'[^\d]+', '', edit_text)
            # (lb): Alright, maybe this parsing should be pushed down to parse_dated.
            # - Add punctuation based on number of digits, i.e.,
            #   YYYY 4 | YYYYMM 6 | YYYYMMDD 8 | YYYYMMDDhh 10 | YYYYMMDDhhmm 12
            # - Note that clock time still works, e.g., `1000` will be interpreted
            #   as 10 AM. And `100` as 1 AM.
            pattern = r''
            replace = r''
            n_digits = len(numberless)
            if n_digits >= 4:
                pattern += r'(....)'
                replace += r'\1'
            if n_digits >= 6:
                pattern += r'(..)'
                replace += r'-\2'
            if n_digits >= 8:
                pattern += r'(..)'
                replace += r'-\3'
            if n_digits >= 10:
                pattern += r'(..)'
                replace += r' \4'
            if n_digits >= 12:
                pattern += r'(..)'
                replace += r':\5'
            reformatted = re.sub(pattern, replace, numberless)
            self.debug('reformatted: {}'.format(reformatted))
            return reformatted

        return _parse_dated()

    # ***

    @catch_action_exception
    # SKIP: @ZoneContent.Decorators.reset_showing_help
    def toggle_focus_description(self, event):
        self.zone_manager.toggle_focus_time_widget(self.zone_manager.content_control)

    # ***

    @catch_action_exception
    def undo_command(self, event):
        # FIXME/2019-01-21: Is this correct? Use case:
        #   - Press 'e'; edit 'end'; press Enter to apply change.
        #   - Press 'e' again; edit 'end'; press Ctrl-z.
        #   - Expect: previous (intermediate) end time, not original end time!
        #   - For original end time: press 'R' reset!
        orig_fact = self.zone_manager.facts_diff.orig_fact
        # FIXME/2019-01-14 18:11: Localization/l10n/timezone'ation...
        #                           start_fmt_local vs start_fmt_utc, and end...
        # When editing, reset widget to unedited time (do not go through undo stack).
        if self.active_widgets is self.widgets_start:
            event.current_buffer.text = orig_fact.start_fmt_local
        elif self.active_widgets is self.widgets_end:
            event.current_buffer.text = orig_fact.end_fmt_local

    @catch_action_exception
    def redo_command(self, event):
        # We could restore the edited time that the user undid.
        # But there's not much utility in that.
        pass




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""FactsManager_RiftInc"""

__all__ = (
    'FactsManager_RiftInc',
)


class FactsManager_RiftInc(object):
    """"""
    def jump_rift_inc(self):
        next_fact = self.find_rift_fact(is_next=True)
        if next_fact is None:
            return self.jump_to_latest_fact(reason='rift-inc')
        return next_fact

    def jump_fact_final(self, reason='fact-final'):
        """"""
        return self.jump_to_latest_fact(reason, include_edge_gap=True)

    # ***

    def jump_to_latest_fact(self, reason, include_edge_gap=False):
        """"""
        def _jump_to_latest_fact():
            final_group = self.groups[-1] if include_edge_gap else ceil_groups()
            _final_group, final_fact = group_latest(final_group)
            next_fact = final_fact
            # Set Edit/FactsManagers' curr_fact; Clear _jump_time_ref.;
            # Update viewed_fact_pks; Maybe call reset_paste/reset paste_cnt.
            self.fulfill_jump(next_fact, reason=reason)
            if include_edge_gap and next_fact.end is not None:
                # Create the active gap Fact.
                # - This'll happen if final Fact is complete: start dob (starts
                #   on final Fact), press 'left' 0 or more times, press 'G'.
                next_fact = self.jump_fact_inc()
                self.controller.affirm(next_fact.is_gap)
            return next_fact

        def ceil_groups():
            if (
                (self.curr_group is self.groups[0])
                and (self.curr_index < (len(self.groups[0]) - 1))
            ):
                # Looking at new, prev Facts, and not the final new
                # Fact, so scroll forward to the last new, prev Fact.
                final_group = self.groups[0]
            else:
                final_group = self.groups[-1]
                # If new, next Facts, scroll to latest Fact, then to final
                # new, next Fact.
                if (
                    (len(self.groups) > 1)
                    and (
                        (self.curr_group is not self.groups[-1])
                        and (
                            (self.curr_group is not self.groups[-2])
                            or (self.curr_index < (len(self.groups[-2]) - 1))
                        )
                    )
                ):
                    final_group = self.groups[-2]
            return final_group

        def group_latest(final_group):
            final_fact = final_group[-1]
            if final_group is not self.groups[-1]:
                return final_group, final_fact
            # else, we're on the final group.
            elif final_group.until_time_stops:
                # The final_group's last element is the active Fact.
                # If it's a gap Fact, however, which is unsaved (contains
                # no user data), caller can opt to receive last real Fact.
                if not include_edge_gap and final_fact.is_gap:
                    # final_fact.is_gap, so try the second-to-last Fact.
                    # - This'll happen if last saved Fact is complete: start dob,
                    # press 'right' to create Gap fact, 'left', 'left', press 'f'
                    # to return to final true fact (find_rift_fact handles it),
                    # then press 'f' again, and the code will branch to here.
                    try:
                        final_fact = final_group[-2]
                    except IndexError:
                        # Well, at least we tried.
                        # (lb): I need to test this branch. Happens on empty db.
                        pass
                return final_group, final_fact
            self.controller.affirm(final_fact.next_fact is None)
            latest_fact = self.controller.find_latest_fact()
            if not latest_fact:
                # Empty database, meaning local Facts unsaved.
                self.controller.affirm(final_fact.unstored)
            else:
                try:
                    latest_fact = self.by_pk[latest_fact.pk]
                    self.controller.affirm(latest_fact.orig_fact is not None)
                except KeyError:
                    # The latest Fact from the db is new to us!
                    self.controller.affirm(latest_fact.orig_fact is None)
                    latest_fact.orig_fact = 0
                    self.add_facts([latest_fact])
            if (
                (not latest_fact)
                or (latest_fact.deleted)
                or (latest_fact.pk == final_fact.pk)
                or (latest_fact < final_fact)
            ):
                # The call above to find_latest_fact excluded deleted, so only
                # way latest_fact.deleted is if local copy in self.by_pk is so.
                return final_group, final_fact
            # If the new_facts were after latest_fact, we'll have loaded
            # latest_fact, but we'll be showing the final new, next Fact.
            final_group = self.groups[-1]
            latest_fact = final_group[-1]
            return final_group, latest_fact

        return _jump_to_latest_fact()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel"""

from gettext import gettext as _

import asyncio
import time

import click_hotoffthehamster as click

from inflector import English, Inflector

from nark.helpers.dev.profiling import profile_elapsed

from dob_bright.styling.style_conf import color as styling_color
from dob_bright.styling.style_engine import StyleEngine
from dob_bright.termio.errors import dob_been_warned_reset

from dob_prompt.prompters.triple_prompter import ask_user_for_edits

from dob_viewer.ptkui.re_confirm import confirm

from ..ptkui.dialog_overlay import show_message

from .action_manager import ActionManager
from .edits_manager import EditsManager
from .exceptions import catch_action_exception
from .update_handler import UpdateHandler
from .zone_content import ZoneContent
from .zone_manager import ZoneManager

__all__ = (
    'Carousel',
)


class Carousel(object):
    """"""
    def __init__(
        self,
        controller,
        edit_facts,
        orig_facts,
        dirty_callback,
        dry,
        content_lexer=None,  # Lexer() instance, used by ZoneContent.
        style_classes=None,  # Style dict, used by all the zone_*.py.
        rules_confobj=None,  # Matching logic and container-class map.
        no_completion=None,  # act/cat/tag list to not complete/suggest.
    ):
        self.controller = controller
        self.edits_manager = EditsManager(
            controller,
            edit_facts=edit_facts,
            orig_facts=orig_facts,
            dirty_callback=dirty_callback,
            error_callback=self.error_callback,
        )
        self.dry = dry
        self.content_lexer = content_lexer
        self.setup_styling(style_classes, rules_confobj)
        self.no_completion = no_completion
        self.action_manager = ActionManager(self)
        self.update_handler = UpdateHandler(self)
        # We'll set up the ZoneManager each time we use the event_loop.
        self.zone_manager = None
        self._avail_width = None
        self.setup_async()

    # ***

    @property
    def carousel(self):
        """Magic for reset_showing_help."""
        return self

    # ***

    @property
    def avail_width(self, force=False):
        if force or self._avail_width is None:
            self._avail_width = self.calculate_available_width()
        return self._avail_width

    def calculate_available_width(self):
        # NOTE: Without the - 2, to account for the '|' borders,
        #       app apparently hangs.
        full_width = click.get_terminal_size()[0] - 2
        if not self.style_classes['content-width']:
            return full_width
        avail_width = min(self.style_classes['content-width'], full_width)
        return avail_width

    # ***

    def setup_async(self):
        # 2019-01-25: (lb): We run asynchronously to support such features as
        # tick_tock_now, which keeps the active fact's end clock time updated.
        # However, working with asyncio can be somewhat tricky. So, should you
        # need to disable it, here's a switch.
        # 2020-01-28: PTK3 is always-async, FYI.
        # - But we still need to be able to disable asyncio for tests.
        self._async_enable = True
        self._confirm_exit = False

    @property
    def async_enable(self):
        assert self._async_enable  # 2020-02-01: Always True now because PTK3.
        return self._async_enable

    @property
    def confirm_exit(self):
        return self._confirm_exit

    @confirm_exit.setter
    def confirm_exit(self, confirm_exit):
        self._confirm_exit = confirm_exit

    # ***

    def setup_styling(self, style_classes, rules_confobj):
        """"""

        def _setup_styling():
            setup_style_classes(style_classes)
            setup_rules_confobj(rules_confobj)

        def setup_style_classes(style_classes):
            if style_classes is None:
                # HARDCODED/DEFAULT: style_classes default: 'color'.
                # (lb): The styling_color() method (an alias of various_style.color())
                # qualifies as a *hardcoded* value (e.g., we could instead call light()
                # or night()). But this is merely a fallback in case the user cleared
                # the setting in their configuration, and then for some reason our
                # load_style_classes() stopped fetching the default first. In fact,
                # per that last point, this code is technically unreachable!
                self.controller.affirm(False)  # Nawgunnhappen.
                style_classes = styling_color()
            self.style_classes = style_classes

        def setup_rules_confobj(rules_confobj):
            self.style_engine = StyleEngine(rules_confobj)

        _setup_styling()

    def process_style_rules(self, ppt_widget, friendly_name, fact=None):
        """Apply custom user classes from ~/.config/dob/styling/styles|rules.conf
        to, e.g., use custom color backgrounds for Facts with matching Category."""
        fact = fact or self.carousel.edits_manager.curr_edit
        return self.style_engine.process_style_rules(ppt_widget, friendly_name, fact)

    # ***

    def gallop(self, **kwargs):
        self.standup_once()
        if self.async_enable:
            # Get the OS thread's event loop.
            self.event_loop = asyncio.get_event_loop()
        confirmed_facts = self.run_edit_loop(**kwargs)

        # (lb): We did not start the event loop, so we should not stop it, e.g.,:
        #     self.async_enable and self.event_loop and self.event_loop.stop()
        # which I mention only because I generally like to clean up after myself.
        # (Nonetheless, it's harmless if the application is exiting anyway.
        #  It merely sets a boolean: `event_loop._stopping = True`.)

        # CLOSED_LOOP: (lb): Remember duration until self.event_loop.close()
        # called. At least until we figure out how to wait() to avoid kludge.
        #
        # FIXME/2019-01-27: Disable close_countdown and find better solution.
        # (And leaving this code-comment until that solution is discovered.)
        #
        #     if self.async_enable:
        #         self.close_countdown = time.time()

        return self.edits_manager.prepared_facts if confirmed_facts else []

    def whoa_nellie(self):
        """"""
        # CLOSED_LOOP: (lb): If we do not clean up properly (or fake it, and
        # drag our feet after quitting the Carousel, but before exiting app),
        # you might experience a RuntimeError such as:
        #
        #     Exception in thread Thread-12:
        #     Traceback (most recent call last):
        #       File "/usr/lib/python3.6/threading.py", line 916,
        #         in _bootstrap_inner
        #           self.run()
        #       ...
        #       File "/usr/lib/python3.6/asyncio/base_events.py", line 366,
        #         in _check_closed
        #           raise RuntimeError('Event loop is closed')
        #     RuntimeError: Event loop is closed
        #
        # 2019-01-27: I think this is reproducible on a large dob-import, after
        # pressing <Ctrl-s> to exit the Carousel and beginning persisting Facts.
        #
        # My first guess was the tick-tock task, but there are 5 threads' stack
        # traces (like the example reprinted above) printed to the terminal.
        # So maybe PPT has some other input processing or what not going on.
        #
        # (Also, note that we bounce around between 3 different uses of PPT,
        # stopping one to start another. 2 uses of PPT -- editing the act@gory
        # and editing tags -- are run without the event loop, whereas the other
        # use, running the Carousel, uses the event loop. There could be some
        # clean up or waiting not happening between uses.)
        #
        # I tried calling other functions (based on suggestions I found online), e.g.,
        #
        #     self.event_loop.run_until_complete(very_long_coroutine())
        #     ...
        #     self.event_loop.stop()
        #     self.event_loop.run_forever()
        #     self.event_loop.close()
        #
        # but none of that helped.
        #
        # (lb): I've seen the error with a ½ sec. wait; but never with 1 second.
        #   But there's no way I'm imposing 1 sec. wait. We'll find another way.
        #
        #   NAH:
        #
        #     wait_at_least_secs = 1.0
        #     remaining = wait_at_least_secs - (time.time() - self.close_countdown)
        #     if remaining > 0:
        #         time.sleep(remaining)
        #         pass

    @property
    def prepared_facts(self):
        return self.edits_manager.prepared_facts

    # ***

    def run_edit_loop(self, **kwargs):
        confirmed_facts = False
        used_prompt = None
        self.enduring_edit = True
        while self.enduring_edit:
            self.runloop(**kwargs)
            if not self.confirm_exit:
                if self.enduring_edit:
                    used_prompt = self.prompt_fact_edits(used_prompt)
                elif not self.edits_manager.user_viewed_all_new_facts:
                    confirmed = self.process_save_early()
                    if confirmed:
                        confirmed_facts = True
                    else:
                        self.enduring_edit = True  # Keep looping.
            if self.confirm_exit:
                confirmed = self.process_exit_request()
                if not confirmed:
                    self.enduring_edit = True  # Keep looping.
            elif not self.enduring_edit:
                confirmed_facts = True  # All done; user looked at all Facts.

            # CPR_ISSUE: An ever-so-brief pause after editing, before returning
            # to the Carousel, apparently precludes the Cursor Position Request
            # problem (which is documented in a comment in runloop_async).
            # 2019-01-27: (lb): I added this sleep before I added standup_always.
            #   MAYBE/2019-01-27: (lb): TEST: Determine if this sleep still
            #   necessary. Or just leave it, it's painless.
            if self.enduring_edit:
                time.sleep(0.01)

        return confirmed_facts

    def process_exit_request(self):
        if not self.edits_manager.is_dirty:
            # No Facts edited.
            return True
        question = _('\nReally exit without saving?')
        allow_mash_quit = self.controller.config['dev.allow_mash_quit']
        try:
            confirmed = confirm(
                question,
                erase_when_done=True,
                allow_mash_quit=allow_mash_quit,
            )
        except KeyboardInterrupt:
            confirmed = False
        return confirmed

    def process_save_early(self):
        question = _('\nReally save without verifying all Facts?')
        confirmed = confirm(question, erase_when_done=True)
        return confirmed

    def prompt_fact_edits(self, used_prompt):
        try:
            used_prompt = self.user_prompt_edit_fact(used_prompt)
        except KeyboardInterrupt:
            # Ye olde Ctrl-c, and not an Exception.
            self.enduring_edit = False
            self.confirm_exit = True
        else:
            self.zone_manager.rebuild_containers()
        return used_prompt

    def user_prompt_edit_fact(self, used_prompt):
        edit_fact = self.edits_manager.undoable_editable_fact(what='prompt-user')
        used_prompt = self.prompt_user(edit_fact, used_prompt)
        self.edits_manager.apply_edits(edit_fact)
        self.zone_manager.reset_diff_fact()
        return used_prompt

    def prompt_user(self, edit_fact, used_prompt):
        used_prompt = ask_user_for_edits(
            self.controller,
            edit_fact,
            always_ask=True,
            prompt_agent=used_prompt,
            restrict_edit=self.restrict_edit,
            no_completion=self.no_completion,
        )
        return used_prompt

    # ***

    def standup_once(self):
        self.edits_manager.stand_up()

    def standup_always(self, **kwargs):
        self.managers_standup()
        return self.check_nak_if_errors_or_build_and_show(**kwargs)

    def managers_standup(self):
        self.zone_manager = ZoneManager(self)
        self.zone_manager.standup()
        self.action_manager.standup()
        self.update_handler.standup()
        self.action_manager.finalize_standup()

    def check_nak_if_errors_or_build_and_show(self, **kwargs):
        # Linger for errors, with basic config, style config, key binding,
        # etc., otherwise Carousel overwrites screen and user won't see them.
        if self.pause_on_error_message_maybe():
            self.zone_manager.build_and_show(**kwargs)
            return True
        return False

    def pause_on_error_message_maybe(self):
        if not dob_been_warned_reset():
            return True
        # (lb): My first approach was to sleep:
        #   time.sleep(2.666)
        # but perhaps requiring acknowledgement is better, because even after
        # quitting editor, error is still not visible unless the user scrolls.
        # - There's a Linux-only single-key reader, i.e., so user needs not ENTER,
        #     https://pypi.org/project/readchar/
        #   or, for cross-platform solution, see:
        #     https://stackoverflow.com/questions/510357/
        #       python-read-a-single-character-from-the-user
        #   but that seems like a lot of extra work for error handler.
        click.echo(_('Press ENTER to acknowledge.'))
        # Be sure to gobble more than user types, lest it goes to PTK input.
        stdin_text = click.get_text_stream('stdin')
        try:
            stdin_text.readline()
        except KeyboardInterrupt:
            click.echo(_('You got it!'))
            return False
        return True

    # ***

    def runloop(self, **kwargs):
        self.confirm_exit = False
        # Use enduring_edit as a trinary to know if the Carousel's run-loop
        # completes unexpectedly (it'll be None), or if the user interacted
        # with the application and it exited its loop deliberately (it'll be
        # True or False).
        self.enduring_edit = None
        self.restrict_edit = ''
        rerun_cnt = 0
        keep_running = True
        # MAGIC_NUMBER: CPR_ISSUE: Do not rerun > once, lest stuck in feedback loop.
        while keep_running and rerun_cnt < 2:
            keep_running = self.runloop_run(**kwargs)
            rerun_cnt += 1

    def runloop_run(self, **kwargs):
        profile_elapsed('To dob runloop')

        # CPR_ISSUE: (lb): 2019-01-27: This might be the Ultimate Fix, by which
        # I mean I added this code last, and it might all that's needed to get
        # around the CPR issue (in which case, the `enduring_edit is None`
        # kludge below may be unnecessary).
        if not self.standup_always(**kwargs):
            return False

        if self.async_enable:
            rerun = self.runloop_async()
            return rerun
        else:
            assert False  # 2020-02-01: Dead path now because always async.
            self.zone_manager.application.run()
            return False

    def runloop_async(self):
        # CPR_ISSUE: (lb): A Funky Business upon Rerunning Carousel.
        #
        # After the user edits the description, or the act@gory, or tags, the
        # application returns to the Carousel. But sometimes the Carousel
        # completes immediately. Then our code, not seeing enduring_edit,
        # exits, though the user was expecting to see the Carousel again.
        #
        # The user also sees what looks like CPR (cursor position request)
        # input on the terminal, for instance,
        #
        #     ^[[12;1R>
        #
        # I cannot deterministically reproduce the behavior other than to
        # cycle between the Carousel and editing, mashing keys to jump
        # quickly between states, and occasionally saving.
        #
        # We can kludge around the behavior by rerunning the Carousel if
        # it completely quickly (i.e., in less time then then use could have
        # interacted with it). (Though this introduces additional pitfalls,
        # like falling into an infinite feedback loop. Just be careful!)
        rerun = False

        # 2020-03-31: (lb): In earlier PTK, calling run_async() would draw the
        # app, but not run its event loop (so some Carousel code would run).
        # But now, this is truly async and nothing is invoked, just a coroutine
        # is returned.
        run_async = self.zone_manager.application.run_async()

        # Create the tick-tock task.
        # MAYBE/2020-01-30: See (new?) in PTK3: create_background_task.
        # - Maybe have PTK3 manage tick-tock, not us. (This works fine, though).
        tck_task = self.event_loop.create_task(self.tick_tock_now())

        # Leave tck_task out of tasks and manage separately.
        tasks = [run_async, ]

        # Run the Carousel!
        self.event_loop.run_until_complete(asyncio.wait(tasks))

        # Check if the Carousel exited deliberately or not.
        if self.enduring_edit is None:
            # CPR_ISSUE: (lb): Look in PPT for ask_for_cpr: this sends the CPR:
            #   Query Cursor Position: <ESC>[6n
            # The other junk you see on the terminal are cursor positions:
            #   Report Cursor Position: <ESC>[{ROW};{COLUMN}R
            # If code gets stuck in a loop, the request acts like Schrödinger's
            # cat and changes value each time it prints to the terminal, e.g.,
            #   ^[[11;1R^[[11;9R^[[11;17R^[[11;26R^[[11;35R
            #           ^[[12;9R^[[12;17R^...
            # (In other cases, you might see `;226R;226R;226R...` instead).
            # Ref:
            #   prompt_toolkit/renderer.py
            #       request_absolute_cursor_position
            self.controller.client_logger.warning('KLUDGE! Re-running Carousel.')
            rerun = True

        tck_task.cancel()
        # Similar to `await tck_task`, but this method not async function, so
        # go through the event loop. -In past, (lb)'s seen warning if skipped:
        #   RuntimeWarning: coroutine 'wait' was never    awaited
        self.event_loop.run_until_complete(asyncio.wait([tck_task, ]))

        return rerun

    # ***

    async def tick_tock_now(self):
        """"""
        async def _tick_tock_now():
            tocking = True
            while tocking:
                tocking = await tick_tock_loop()

        async def tick_tock_loop():
            if not await sleep_to_refresh():
                return False
            refresh_viewable()
            return True

        async def sleep_to_refresh():
            try:
                # (lb): I tried a few different behaviors here, e.g.,
                # longer sleep, or even only updating if only so much
                # time has passed (i.e., because we only need to update
                # the "now" time, we only need to really redraw every
                # second), but the seconds of the "now" time would visibly
                # increment unevenly. I settled around 50 msecs. and it
                # doesn't seem to make the user interaction at all sluggish.
                # ... Hahaha, whelp, 50 msecs. makes 1 CPU run 100% hot.
                # And 500 msecs. make 1 CPU run 20%, instead of 10....
                # But if sleep is ⅔ secs., seconds increment is jumpy.
                # 500 msecs. seems to work well.
                await asyncio.sleep(0.5)
            except asyncio.CancelledError:
                return False
            except Exception as err:
                self.controller.client_logger.warning(
                    "Unexpected async err: {}".format(err),
                )
            return True

        def refresh_viewable():
            self.zone_manager.selectively_refresh()
            self.zone_manager.application.invalidate()

        await _tick_tock_now()

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def exit_command(self, event):
        """"""
        self.confirm_exit = True
        self.enduring_edit = False
        event.app.exit()

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def exit_quietly(self, event):
        """"""
        # (lb): A little awkward: Meta-key combinations start with 'escape',
        # even though the user doesn't explicitly press the escape key. E.g.,
        # if you press ESC, PPT waits a moment, and if you don't press another
        # key, it invokes the callback for ESC; but if you press an escaped
        # combination that does not have a callback, e.g., if ALT-LEFT is not
        # bound, and the user presses ALT-LEFT, then the handler for ESC is
        # called (because the ALT-LEFT combination starts with an escape).
        if len(event.key_sequence) != 1:
            return
        kseq = event.key_sequence[0]
        if kseq.key == 'escape' and kseq.data != '\x1b':
            return
        # Allow easy 'q' cancel on editing existing Facts.
        # FIXME: Should allow for new_facts, too... but we don't track edited
        #        (though we do in each Fact's dirty_reasons; but we're
        #        consolidating behavior in new class, so can wait to fix).
        #        Then again, how much do I care? Import should be rare.
        if not self.edits_manager.is_dirty:
            self.confirm_exit = True
            self.enduring_edit = False
            event.app.exit()

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def save_edited_and_exit(self, event):
        self.save_edited_and_live(event)
        # (lb): Exit Carousel, then Save. Traditional Import behavior
        # (before running save/save_edited_and_live was implemented).
        self.enduring_edit = False
        event.app.exit()
        return

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def save_edited_and_live(self, event):
        """"""
        # MAYBE/TESTME/2019-02-01: (lb): If running save on dob-import, use progger?
        # - Import save can take a while (because checks for conflicts on each Fact)
        #   -- do we need a Carousel progress display (progger)?
        curr_fact, saved_facts = self.edits_manager.save_edited_facts()
        if saved_facts is None:
            # Indicates error during save, and error message was displayed.
            return
        if not saved_facts:
            curr_fact = None
        if saved_facts:
            self.controller.post_process(
                self.controller,
                saved_facts,
                show_plugin_error=self.show_plugin_error,
                carousel_active=True,
            )
        edit_cnt = len(saved_facts)
        self.zone_manager.finalize_jump(
            curr_fact,
            noop_msg=_('Nothing to save'),
            jump_msg=_('Saved {} {}'.format(
                edit_cnt,
                Inflector(English).conditional_plural(edit_cnt, 'fact'),
            )),
        )

    # ***

    def error_callback(self, errmsg):
        show_message(
            self.zone_manager.root,
            _('Wah wah'),
            _("dob is buggy! {0}").format(errmsg),
        )

    def show_plugin_error(self, errmsg):
        show_message(
            self.zone_manager.root,
            _('Oops!'),
            _('{0}').format(errmsg),
        )

    # ***

    def dev_breakpoint(self, event):
        if not self.controller.config['dev.catch_errors']:
            self.controller.client_logger.warning(
                'Please enable ‘dev.catch_errors’ to use live debugging.'
            )
            return
        self.pdb_set_trace(event)

    def pdb_set_trace(self, event):
        import pdb
        # Just some convenience variables for the developer.
        # F841: local variable '...' is assigned to but never used
        edits = self.edits_manager  # noqa: F841
        facts = self.edits_manager.conjoined  # noqa: F841
        groups = self.edits_manager.conjoined.groups  # noqa: F841
        # Reset terminal I/O to fix interactive pdb stdin echo.
        self.pdb_break_enter()
        pdb.set_trace()
        pass  # Poke around; print variables; then [c]ontinue.
        self.pdb_break_leave(event)

    def pdb_break_enter(self):
        self.controller.pdb_break_enter()

    def pdb_break_leave(self, event=None):
        self.controller.pdb_break_leave()
        # Redraw everything. But don't both with invalidate, e.g.,:
        #   self.carousel.zone_manager.application.invalidate()
        # but rather find the renderer and clear that.
        # This'll also reset the cursor, so nice!
        self.controller.affirm(
            (event is None) or (event.app is self.zone_manager.application),
        )
        self.zone_manager.application.renderer.clear()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""FactsManager_RiftDec"""

__all__ = (
    'FactsManager_RiftDec',
)


class FactsManager_RiftDec(object):
    """"""
    def jump_rift_dec(self):
        """"""
        # find_rift_fact calls jump_to_fact_nearest (which calls fulfill_jump).
        prev_fact = self.find_rift_fact(is_prev=True)
        if prev_fact is None:
            return self.jump_to_oldest_fact(reason='rift-dec')
        return prev_fact

    def jump_fact_first(self, reason='fact-first'):
        """"""
        return self.jump_to_oldest_fact(reason, include_edge_gap=True)

    # ***

    def jump_to_oldest_fact(self, reason, include_edge_gap=False):
        """"""
        def _jump_to_oldest_fact():
            first_group = self.groups[0] if include_edge_gap else floor_groups()
            _first_group, first_fact = group_oldest(first_group)
            prev_fact = first_fact
            self.fulfill_jump(prev_fact, reason=reason)
            if include_edge_gap and not prev_fact.is_gap:
                # Create the prehistoric gap Fact. Run dob then `gg` to test.
                prev_fact = self.jump_fact_dec()
                self.controller.affirm(prev_fact.is_gap)
            return prev_fact

        def floor_groups():
            if (
                (self.curr_group is self.groups[-1])
                and (self.curr_index > 0)
            ):
                # Looking at new, next Facts, and not the first new
                # Fact, so scroll back to the first new, next Fact.
                first_group = self.groups[-1]
            else:
                first_group = self.groups[0]
                # If new, prev Facts, scroll to oldest Fact, then to first
                # new, prev Fact. (Note: If there are new, prev Facts, the
                # oldest Fact is already established at self.groups[1][0].)
                if (
                    (self.curr_group is not self.groups[0])
                    and (
                        (self.curr_group is not self.groups[1])
                        or (self.curr_index > 0)
                    )
                ):
                    first_group = self.groups[1]
            return first_group

        def group_oldest(first_group):
            first_fact = first_group[0]
            if first_group is not self.groups[0]:
                return first_group, first_fact
            # else, we're on the first group.
            elif first_group.since_time_began:
                # On first group, which starts with gap Fact.
                if not include_edge_gap and first_fact.is_gap:
                    # Try for first actual, saved Fact, if one exists. This
                    # implements the difference between 'F' and 'gg' commands.
                    try:
                        first_fact = first_group[1]
                    except IndexError:
                        # I.e., database is empty.
                        pass
                return first_group, first_fact
            self.controller.affirm(first_fact.prev_fact is None)
            oldest_fact = self.controller.find_oldest_fact()
            if not oldest_fact:
                self.controller.affirm(first_fact.unstored)
            else:
                try:
                    oldest_fact = self.by_pk[oldest_fact.pk]
                    self.controller.affirm(oldest_fact.orig_fact is not None)
                except KeyError:
                    self.controller.affirm(oldest_fact.orig_fact is None)
                    oldest_fact.orig_fact = 0
                    self.add_facts([oldest_fact])
            if (
                (not oldest_fact)
                or (oldest_fact.deleted)
                or (oldest_fact.pk == first_fact.pk)
                or (oldest_fact > first_fact)
            ):
                return first_group, first_fact
            # If the new_facts were before oldest_fact, we'll have loaded
            # oldest_fact, but we'll be showing the first new, prev Fact.
            first_group = self.groups[0]
            first_fact = first_group[0]
            return first_group, first_fact

        return _jump_to_oldest_fact()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import pytest

from prompt_toolkit.input.defaults import create_pipe_input
from prompt_toolkit.output import DummyOutput

from dob_bright.crud.parse_input import parse_input

from dob_viewer.ptkui import re_confirm
from dob_viewer.traverser.save_confirmer import prompt_and_save_confirmer


@pytest.fixture
def new_facts(controller_with_logging):
    input_stream = open(IMPORT_PATH, 'r')
    new_facts = parse_input(
        controller_with_logging,
        file_in=input_stream,
        progress=None,
    )
    return new_facts


IMPORT_PATH = './tests/fixtures/test-import-fixture.rst'
"""Path to the import file fixture, which is full of Factoids."""


class TestBasicCarousel(object):
    """Non-interactive Interactive Carousel tests."""

    # ***

    def _feed_cli_with_input(
        self,
        controller_with_logging,
        new_facts,
        key_sequence,
        mocker,
    ):
        # (lb): In the original tests, back when PTK2, we monkeypatch'ed sys,stdin
        # to a pipe opened from a pty.openpty pseudo terminal, which smelled very
        # fragile, and came with the caveat that Python's pty library probably
        # doesn't run well on non-Linux. Thankfully, PTK3 has a formal mechanism
        # for overriding I/O. See decent examples under prompt_toolkit/tests/.

        # The Carousel will prompt for confirmation on quit without having saved.
        mocker.patch.object(re_confirm, 'confirm', return_value=True)

        inp = create_pipe_input()
        try:
            inp.send_text(key_sequence)
            # Ignoring return value: saved_facts.
            prompt_and_save_confirmer(
                controller_with_logging,
                edit_facts=new_facts,
                # Test apparatus ('input' and 'output' passed as kwargs to Application).
                input=inp,
                output=DummyOutput(),
            )
        finally:
            inp.close()

    # ***

    @pytest.mark.parametrize(
        ('key_sequence'),
        [
            # *** test_basic_import4_left_arrow_three_time
            #
            # Test left-arrowing and first (early Life) gap fact.
            # Left arrow three times.
            # - First time creates and jumps to gap fact.
            # - Second time causes at-first-fact message.
            # - Third time's a charm.
            [
                '\x1bOD',   # Left arrow ←.
                '\x1bOD',   # Left arrow ←.
                '\x1bOD',   # Left arrow ←.
                '\x11',     # Ctrl-Q.
                '\x11',     # Ctrl-Q.
                '\x11',     # Ctrl-Q.
            ],

            # *** test_basic_import4_right_arrow_left_arrow
            #
            [
                # Arrow right, arrow left.
                '\x1bOD',
                '\x1bOC',
                # Three Cancels don't make a Right.
                '\x11',
                '\x11',
                '\x11',
                # FIXME/2019-02-20: Because, what, arrowing left goes to
                #                   Previous Big Bang Gap Fact,
                #                   so extra Ctrl-Q needed?
                #                   Oddly, in log, I still only see 2 exit_command's!
                #                   But apparently we need 4 strokes to exit.
                '\x11',
            ],

            # *** test_basic_import4_G_go_last
            [
                # Jump to final fact.
                'G',
                '\x11',
                '\x11',
                '\x11',
            ],
        ],
    )
    def test_basic_import4_left_arrow_three_time(
        self,
        controller_with_logging,
        new_facts,
        key_sequence,
        mocker,
    ):
        self._feed_cli_with_input(
            controller_with_logging,
            new_facts,
            ''.join(key_sequence),
            mocker,
        )

